@@ -129,11 +129,6 @@ sync_array_t*	sync_primary_wait_array;
 /* This variable is set to TRUE when sync_init is called */
 ibool	sync_initialized	= FALSE;
 
-/* Global list of database mutexes (not OS mutexes) created. */
-UT_LIST_BASE_NODE_T(mutex_t)	mutex_list;
-
-/* Mutex protecting the mutex_list variable */
-mutex_t		mutex_list_mutex;
 
 typedef struct sync_level_struct	sync_level_t;
 typedef struct sync_thread_struct	sync_thread_t;
@@ -146,6 +141,12 @@ sync_thread_t*	sync_thread_level_arrays;
 /* Mutex protecting sync_thread_level_arrays */
 mutex_t	sync_thread_mutex;
 
+/* Global list of database mutexes (not OS mutexes) created. */
+ut_list_base_node_t  mutex_list;
+
+/* Mutex protecting the mutex_list variable */
+mutex_t mutex_list_mutex;
+
 /* Latching order checks start when this is set TRUE */
 ibool	sync_order_checks_on	= FALSE;
 
@@ -201,6 +202,9 @@ void
 mutex_create_func(
 /*==============*/
 	mutex_t*	mutex,		/* in: pointer to memory */
+#ifdef UNIV_DEBUG
+	const char*	cmutex_name,	/* in: mutex name */
+#endif /* UNIV_DEBUG */
 	const char*	cfile_name,	/* in: file name where created */
 	ulint		cline)		/* in: file line where created */
 {
@@ -210,6 +214,7 @@ mutex_create_func(
 	os_fast_mutex_init(&(mutex->os_fast_mutex));
 	mutex->lock_word = 0;
 #endif
+	mutex->event = os_event_create(NULL);
 	mutex_set_waiters(mutex, 0);
 	mutex->magic_n = MUTEX_MAGIC_N;
 #ifdef UNIV_SYNC_DEBUG
@@ -219,6 +224,19 @@ mutex_create_func(
 	mutex->level = SYNC_LEVEL_NONE;
 	mutex->cfile_name = cfile_name;
 	mutex->cline = cline;
+#ifndef UNIV_HOTBACKUP
+	mutex->count_os_wait = 0;
+# ifdef UNIV_DEBUG
+  mutex->cmutex_name=     cmutex_name;
+  mutex->count_using=     0;
+  mutex->mutex_type=      0;
+  mutex->lspent_time=     0;
+  mutex->lmax_spent_time=     0;
+  mutex->count_spin_loop= 0;
+  mutex->count_spin_rounds=   0; 
+  mutex->count_os_yield=  0;
+# endif /* UNIV_DEBUG */
+#endif /* !UNIV_HOTBACKUP */
 	
 	/* Check that lock_word is aligned; this is important on Intel */
 	ut_ad(((ulint)(&(mutex->lock_word))) % 4 == 0);
@@ -275,6 +293,8 @@ mutex_free(
 		mutex_exit(&mutex_list_mutex);
 	}
 
+	os_event_free(mutex->event);
+
 #if !defined(_WIN32) || !defined(UNIV_CAN_USE_X86_ASSEMBLER) 
 	os_fast_mutex_free(&(mutex->os_fast_mutex));
 #endif
@@ -355,135 +375,190 @@ for the mutex before suspending the thread. */
 void
 mutex_spin_wait(
 /*============*/
-        mutex_t*	   mutex,     	/* in: pointer to mutex */
-	const char*	   file_name, 	/* in: file name where
-					mutex requested */
-	ulint		   line)	/* in: line where requested */
+  mutex_t*  mutex,      /* in: pointer to mutex */
+  const char*    file_name,   /* in: file name where
+                             mutex requested */
+  ulint line) /* in: line where requested */
 {
-        ulint    index; /* index of the reserved wait cell */
-        ulint    i;   	/* spin round count */
-        
-        ut_ad(mutex);
+  ulint    index; /* index of the reserved wait cell */
+  ulint    i;     /* spin round count */
+#if defined UNIV_DEBUG && !defined UNIV_HOTBACKUP
+  ib_longlong lstart_time = 0, lfinish_time; /* for timing os_wait */
+  ulint ltime_diff;
+  ulint sec;
+  ulint ms;
+  uint timer_started = 0;
+#endif /* UNIV_DEBUG && !UNIV_HOTBACKUP */
+  ut_ad(mutex);
 
 mutex_loop:
 
-        i = 0;
+  i = 0;
+
+/* Spin waiting for the lock word to become zero. Note that we do not
+  have to assume that the read access to the lock word is atomic, as the
+  actual locking is always committed with atomic test-and-set. In
+  reality, however, all processors probably have an atomic read of a
+  memory word. */
 
-        /* Spin waiting for the lock word to become zero. Note that we do not
-	have to assume that the read access to the lock word is atomic, as the
-	actual locking is always committed with atomic test-and-set. In
-	reality, however, all processors probably have an atomic read of a
-	memory word. */
-        
 spin_loop:
-	mutex_spin_wait_count++;
+#if defined UNIV_DEBUG && !defined UNIV_HOTBACKUP
+  mutex_spin_wait_count++;
+  mutex->count_spin_loop++;
+#endif /* UNIV_DEBUG && !UNIV_HOTBACKUP */
 
-        while (mutex_get_lock_word(mutex) != 0 && i < SYNC_SPIN_ROUNDS) {
+  while (mutex_get_lock_word(mutex) != 0 && i < SYNC_SPIN_ROUNDS)
+  {
+    if (srv_spin_wait_delay)
+    {
+      ut_delay(ut_rnd_interval(0, srv_spin_wait_delay));
+    }
 
-        	if (srv_spin_wait_delay) {
-        		ut_delay(ut_rnd_interval(0, srv_spin_wait_delay));
-        	}
-        
-             	i++;
-        }
+    i++;
+  }
 
-	if (i == SYNC_SPIN_ROUNDS) {
-		os_thread_yield();
-	}
+  if (i == SYNC_SPIN_ROUNDS)
+  {
+#if defined UNIV_DEBUG && !defined UNIV_HOTBACKUP
+    mutex->count_os_yield++;
+    if (timed_mutexes == 1 && timer_started==0)
+    {
+      ut_usectime(&sec, &ms);
+      lstart_time= (ib_longlong)sec * 1000000 + ms;
+      timer_started = 1;
+    }
+#endif /* UNIV_DEBUG && !UNIV_HOTBACKUP */
+    os_thread_yield();
+  }
 
-	if (srv_print_latch_waits) {
-		fprintf(stderr,
-	"Thread %lu spin wait mutex at %p cfile %s cline %lu rnds %lu\n",
-		(ulong) os_thread_pf(os_thread_get_curr_id()), mutex,
-		mutex->cfile_name, (ulong) mutex->cline, (ulong) i);
-	}
-
-	mutex_spin_round_count += i;
-
-        if (mutex_test_and_set(mutex) == 0) {
-		/* Succeeded! */
-
-#ifdef UNIV_SYNC_DEBUG
-		mutex_set_debug_info(mutex, file_name, line);
+#ifdef UNIV_SRV_PRINT_LATCH_WAITS
+    fprintf(stderr,
+            "Thread %lu spin wait mutex at %p cfile %s cline %lu rnds %lu\n",
+            (ulong) os_thread_pf(os_thread_get_curr_id()), mutex,
+            mutex->cfile_name, (ulong) mutex->cline, (ulong) i);
 #endif
 
-                return;
-   	}
+  mutex_spin_round_count += i;
 
-	/* We may end up with a situation where lock_word is
-	0 but the OS fast mutex is still reserved. On FreeBSD
-	the OS does not seem to schedule a thread which is constantly
-	calling pthread_mutex_trylock (in mutex_test_and_set
-	implementation). Then we could end up spinning here indefinitely.
-	The following 'i++' stops this infinite spin. */
+#if defined UNIV_DEBUG && !defined UNIV_HOTBACKUP
+  mutex->count_spin_rounds += i;
+#endif /* UNIV_DEBUG && !UNIV_HOTBACKUP */
 
-	i++;
-        
-	if (i < SYNC_SPIN_ROUNDS) {
+  if (mutex_test_and_set(mutex) == 0)
+  {
+    /* Succeeded! */
 
-		goto spin_loop;
-	}
-
-        sync_array_reserve_cell(sync_primary_wait_array, mutex,
-        			SYNC_MUTEX,
-				file_name, line,
-				&index);
-
-	mutex_system_call_count++;
-
-	/* The memory order of the array reservation and the change in the
-	waiters field is important: when we suspend a thread, we first
-	reserve the cell and then set waiters field to 1. When threads are
-	released in mutex_exit, the waiters field is first set to zero and
-	then the event is set to the signaled state. */
-        
-	mutex_set_waiters(mutex, 1);
-
-	/* Try to reserve still a few times */
-	for (i = 0; i < 4; i++) {
-            if (mutex_test_and_set(mutex) == 0) {
-
-                /* Succeeded! Free the reserved wait cell */
-
-                sync_array_free_cell(sync_primary_wait_array, index);
-                
 #ifdef UNIV_SYNC_DEBUG
-		mutex_set_debug_info(mutex, file_name, line);
+    mutex_set_debug_info(mutex, file_name, line);
 #endif
 
-		if (srv_print_latch_waits) {
-			fprintf(stderr,
-				"Thread %lu spin wait succeeds at 2:"
-				" mutex at %p\n",
-			(ulong) os_thread_pf(os_thread_get_curr_id()),
-			mutex);
-		}
-		
-                return;
+    goto finish_timing;
+  }
 
-                /* Note that in this case we leave the waiters field
-                set to 1. We cannot reset it to zero, as we do not know
-                if there are other waiters. */
-            }
-        }
+  /* We may end up with a situation where lock_word is
+  0 but the OS fast mutex is still reserved. On FreeBSD
+  the OS does not seem to schedule a thread which is constantly
+  calling pthread_mutex_trylock (in mutex_test_and_set
+  implementation). Then we could end up spinning here indefinitely.
+  The following 'i++' stops this infinite spin. */
 
-        /* Now we know that there has been some thread holding the mutex
-        after the change in the wait array and the waiters field was made.
-	Now there is no risk of infinite wait on the event. */
+  i++;
 
-	if (srv_print_latch_waits) {
-		fprintf(stderr,
-	"Thread %lu OS wait mutex at %p cfile %s cline %lu rnds %lu\n",
-		(ulong) os_thread_pf(os_thread_get_curr_id()), mutex,
-		mutex->cfile_name, (ulong) mutex->cline, (ulong) i);
-	}
-	
-	mutex_system_call_count++;
-	mutex_os_wait_count++;
+  if (i < SYNC_SPIN_ROUNDS)
+  {
+    goto spin_loop;
+  }
 
-        sync_array_wait_event(sync_primary_wait_array, index);
+  sync_array_reserve_cell(sync_primary_wait_array, mutex,
+                          SYNC_MUTEX, file_name, line, &index);
 
-        goto mutex_loop;        
+  mutex_system_call_count++;
+
+  /* The memory order of the array reservation and the change in the
+  waiters field is important: when we suspend a thread, we first
+  reserve the cell and then set waiters field to 1. When threads are
+  released in mutex_exit, the waiters field is first set to zero and
+  then the event is set to the signaled state. */
+
+  mutex_set_waiters(mutex, 1);
+
+  /* Try to reserve still a few times */
+  for (i = 0; i < 4; i++)
+  {
+    if (mutex_test_and_set(mutex) == 0)
+    {
+      /* Succeeded! Free the reserved wait cell */
+
+      sync_array_free_cell(sync_primary_wait_array, index);
+
+#ifdef UNIV_SYNC_DEBUG
+      mutex_set_debug_info(mutex, file_name, line);
+#endif
+
+#ifdef UNIV_SRV_PRINT_LATCH_WAITS
+        fprintf(stderr, "Thread %lu spin wait succeeds at 2:"
+                " mutex at %p\n",
+                (ulong) os_thread_pf(os_thread_get_curr_id()),
+                mutex);
+#endif
+
+      goto finish_timing;
+
+      /* Note that in this case we leave the waiters field
+      set to 1. We cannot reset it to zero, as we do not know
+      if there are other waiters. */
+    }
+  }
+
+  /* Now we know that there has been some thread holding the mutex
+  after the change in the wait array and the waiters field was made.
+Now there is no risk of infinite wait on the event. */
+
+#ifdef UNIV_SRV_PRINT_LATCH_WAITS
+    fprintf(stderr,
+            "Thread %lu OS wait mutex at %p cfile %s cline %lu rnds %lu\n",
+            (ulong) os_thread_pf(os_thread_get_curr_id()), mutex,
+            mutex->cfile_name, (ulong) mutex->cline, (ulong) i);
+#endif
+
+  mutex_system_call_count++;
+  mutex_os_wait_count++;
+
+#ifndef UNIV_HOTBACKUP
+  mutex->count_os_wait++;
+# ifdef UNIV_DEBUG
+  /*
+    !!!!! Sometimes os_wait can be called without  os_thread_yield
+  */
+
+  if (timed_mutexes == 1 && timer_started==0)
+  {
+    ut_usectime(&sec, &ms);
+    lstart_time= (ib_longlong)sec * 1000000 + ms;
+    timer_started = 1;
+  }
+# endif /* UNIV_DEBUG */
+#endif /* !UNIV_HOTBACKUP */
+
+  sync_array_wait_event(sync_primary_wait_array, index);
+  goto mutex_loop;  
+
+finish_timing:
+#if defined UNIV_DEBUG && !defined UNIV_HOTBACKUP
+  if (timed_mutexes == 1 && timer_started==1)
+  {
+    ut_usectime(&sec, &ms);
+    lfinish_time= (ib_longlong)sec * 1000000 + ms;
+
+    ltime_diff= (ulint) (lfinish_time - lstart_time);
+    mutex->lspent_time += ltime_diff;
+    if (mutex->lmax_spent_time < ltime_diff)
+    {
+      mutex->lmax_spent_time= ltime_diff;
+    }
+  }
+#endif /* UNIV_DEBUG && !UNIV_HOTBACKUP */
+  return;
 }
 
 /**********************************************************************
@@ -498,8 +573,8 @@ mutex_signal_object(
 
 	/* The memory order of resetting the waiters field and
 	signaling the object is important. See LEMMA 1 above. */
-
-	sync_array_signal_object(sync_primary_wait_array, mutex);
+	os_event_set(mutex->event);
+	sync_array_object_signalled(sync_primary_wait_array);
 }
 
 #ifdef UNIV_SYNC_DEBUG
@@ -555,6 +630,7 @@ mutex_set_level(
 	mutex->level = level;
 }
 
+
 #ifdef UNIV_SYNC_DEBUG
 /**********************************************************************
 Checks that the current thread owns the mutex. Works only in the debug
@@ -1047,6 +1123,7 @@ sync_thread_add_level(
 		ut_a(sync_thread_levels_g(array, SYNC_PURGE_SYS));
 	} else if (level == SYNC_TREE_NODE) {
 		ut_a(sync_thread_levels_contain(array, SYNC_INDEX_TREE)
+		     || sync_thread_levels_contain(array, SYNC_DICT_OPERATION)
 		     || sync_thread_levels_g(array, SYNC_TREE_NODE - 1));
 	} else if (level == SYNC_TREE_NODE_FROM_HASH) {
 		ut_a(1);
@@ -1075,8 +1152,12 @@ sync_thread_add_level(
 	} else if (level == SYNC_DICT_HEADER) {
 		ut_a(sync_thread_levels_g(array, SYNC_DICT_HEADER));
 	} else if (level == SYNC_DICT) {
+#ifdef UNIV_DEBUG
 		ut_a(buf_debug_prints
 		     || sync_thread_levels_g(array, SYNC_DICT));
+#else /* UNIV_DEBUG */
+		ut_a(sync_thread_levels_g(array, SYNC_DICT));
+#endif /* UNIV_DEBUG */
 	} else {
 		ut_error;
 	}
