@@ -1,9 +1,8 @@
-/* Copyright (C) 2000,2004 MySQL AB & MySQL Finland AB & TCX DataKonsult AB
+/* Copyright (C) 2000-2006 MySQL AB
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
-   the Free Software Foundation; either version 2 of the License, or
-   (at your option) any later version.
+   the Free Software Foundation; version 2 of the License.
 
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
@@ -115,6 +114,26 @@ int ha_heap::close(void)
 }
 
 
+/*
+  Create a copy of this table
+
+  DESCRIPTION
+    Do same as default implementation but use file->s->name instead of 
+    table->s->path. This is needed by Windows where the clone() call sees
+    '/'-delimited path in table->s->path, while ha_peap::open() was called 
+    with '\'-delimited path.
+*/
+
+handler *ha_heap::clone(MEM_ROOT *mem_root)
+{
+  handler *new_handler= get_new_handler(table, mem_root, table->s->db_type);
+  if (new_handler && !new_handler->ha_open(file->s->name, table->db_stat,
+                                           HA_OPEN_IGNORE_IF_LOCKED))
+    return new_handler;
+  return NULL;  /* purecov: inspected */
+}
+
+
 /*
   Compute which keys to use for scanning
 
@@ -176,7 +195,10 @@ int ha_heap::write_row(byte * buf)
   if (table->timestamp_field_type & TIMESTAMP_AUTO_SET_ON_INSERT)
     table->timestamp_field->set_time();
   if (table->next_number_field && buf == table->record[0])
-    update_auto_increment();
+  {
+    if ((res= update_auto_increment()))
+      return res;
+  }
   res= heap_write(file,buf);
   if (!res && (++records_changed*HEAP_STATS_UPDATE_THRESHOLD > 
                file->s->records))
@@ -315,11 +337,11 @@ int ha_heap::rnd_next(byte *buf)
 int ha_heap::rnd_pos(byte * buf, byte *pos)
 {
   int error;
-  HEAP_PTR position;
+  HEAP_PTR heap_position;
   statistic_increment(table->in_use->status_var.ha_read_rnd_count,
 		      &LOCK_status);
-  memcpy_fixed((char*) &position,pos,sizeof(HEAP_PTR));
-  error=heap_rrnd(file, buf, position);
+  memcpy_fixed((char*) &heap_position, pos, sizeof(HEAP_PTR));
+  error=heap_rrnd(file, buf, heap_position);
   table->status=error ? STATUS_NOT_FOUND: 0;
   return error;
 }
@@ -329,21 +351,21 @@ void ha_heap::position(const byte *record)
   *(HEAP_PTR*) ref= heap_position(file);	// Ref is aligned
 }
 
-void ha_heap::info(uint flag)
+int ha_heap::info(uint flag)
 {
-  HEAPINFO info;
-  (void) heap_info(file,&info,flag);
+  HEAPINFO hp_info;
+  (void) heap_info(file,&hp_info,flag);
 
-  records = info.records;
-  deleted = info.deleted;
-  errkey  = info.errkey;
-  mean_rec_length=info.reclength;
-  data_file_length=info.data_length;
-  index_file_length=info.index_length;
-  max_data_file_length= info.max_records* info.reclength;
-  delete_length= info.deleted * info.reclength;
+  records= hp_info.records;
+  deleted= hp_info.deleted;
+  errkey=  hp_info.errkey;
+  mean_rec_length= hp_info.reclength;
+  data_file_length= hp_info.data_length;
+  index_file_length= hp_info.index_length;
+  max_data_file_length= hp_info.max_records* hp_info.reclength;
+  delete_length= hp_info.deleted * hp_info.reclength;
   if (flag & HA_STATUS_AUTO)
-    auto_increment_value= info.auto_increment;
+    auto_increment_value= hp_info.auto_increment;
   /*
     If info() is called for the first time after open(), we will still
     have to update the key statistics. Hoping that a table lock is now
@@ -351,6 +373,7 @@ void ha_heap::info(uint flag)
   */
   if (key_stat_version != file->s->key_stat_version)
     update_key_stats();
+  return 0;
 }
 
 int ha_heap::extra(enum ha_extra_function operation)
@@ -601,7 +624,10 @@ int ha_heap::create(const char *name, TABLE *table_arg,
       seg->length=  (uint) key_part->length;
       seg->flag=    key_part->key_part_flag;
 
-      seg->charset= field->charset();
+      if (field->flags & (ENUM_FLAG | SET_FLAG))
+        seg->charset= &my_charset_bin;
+      else
+        seg->charset= field->charset();
       if (field->null_ptr)
       {
 	seg->null_bit= field->null_bit;
@@ -627,7 +653,7 @@ int ha_heap::create(const char *name, TABLE *table_arg,
   }
   mem_per_row+= MY_ALIGN(share->reclength + 1, sizeof(char*));
   max_rows = (ha_rows) (table->in_use->variables.max_heap_table_size /
-			mem_per_row);
+			(ulonglong) mem_per_row);
   if (table_arg->found_next_number_field)
   {
     keydef[share->next_number_index].flag|= HA_AUTO_KEY;
