@@ -1,9 +1,8 @@
-/* Copyright (C) 2000 MySQL AB & MySQL Finland AB & TCX DataKonsult AB
+/* Copyright (C) 2000-2006 MySQL AB
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
-   the Free Software Foundation; either version 2 of the License, or
-   (at your option) any later version.
+   the Free Software Foundation; version 2 of the License.
 
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
@@ -43,7 +42,7 @@ int mi_create(const char *name,uint keys,MI_KEYDEF *keydefs,
   myf create_flag;
   uint fields,length,max_key_length,packed,pointer,real_length_diff,
        key_length,info_length,key_segs,options,min_key_length_skip,
-       base_pos,varchar_count,long_varchar_count,varchar_length,
+       base_pos,long_varchar_count,varchar_length,
        max_key_block_length,unique_key_parts,fulltext_keys,offset;
   ulong reclength, real_reclength,min_pack_length;
   char filename[FN_REFLEN],linkname[FN_REFLEN], *linkname_ptr;
@@ -59,6 +58,8 @@ int mi_create(const char *name,uint keys,MI_KEYDEF *keydefs,
   my_off_t key_root[MI_MAX_POSSIBLE_KEY],key_del[MI_MAX_KEY_BLOCK_SIZE];
   MI_CREATE_INFO tmp_create_info;
   DBUG_ENTER("mi_create");
+  DBUG_PRINT("enter", ("keys: %u  columns: %u  uniques: %u  flags: %u",
+                      keys, columns, uniques, flags));
 
   if (!ci)
   {
@@ -72,7 +73,6 @@ int mi_create(const char *name,uint keys,MI_KEYDEF *keydefs,
   }
   LINT_INIT(dfile);
   LINT_INIT(file);
-  pthread_mutex_lock(&THR_LOCK_myisam);
   errpos=0;
   options=0;
   bzero((byte*) &share,sizeof(share));
@@ -99,7 +99,7 @@ int mi_create(const char *name,uint keys,MI_KEYDEF *keydefs,
 
 	/* Start by checking fields and field-types used */
 
-  reclength=varchar_count=varchar_length=long_varchar_count=packed=
+  reclength=varchar_length=long_varchar_count=packed=
     min_pack_length=pack_reclength=0;
   for (rec=recinfo, fields=0 ;
        fields != columns ;
@@ -130,14 +130,15 @@ int mi_create(const char *name,uint keys,MI_KEYDEF *keydefs,
       }
       else if (type == FIELD_VARCHAR)
       {
-	varchar_count++;
-	varchar_length+=rec->length-2;
+	varchar_length+= rec->length-1;          /* Used for min_pack_length */
 	packed--;
-	pack_reclength+=1;
-	if (test(rec->length > 257))
-	{					/* May be packed on 3 bytes */
+	pack_reclength++;
+        min_pack_length++;
+        /* We must test for 257 as length includes pack-length */
+        if (test(rec->length >= 257))
+	{
 	  long_varchar_count++;
-	  pack_reclength+=2;
+	  pack_reclength+= 2;			/* May be packed on 3 bytes */
 	}
       }
       else if (type != FIELD_SKIP_ZERO)
@@ -156,6 +157,10 @@ int mi_create(const char *name,uint keys,MI_KEYDEF *keydefs,
       rec--;
       if (rec->type == (int) FIELD_SKIP_ZERO && rec->length == 1)
       {
+        /*
+          NOTE1: here we change a field type FIELD_SKIP_ZERO ->
+          FIELD_NORMAL
+        */
 	rec->type=(int) FIELD_NORMAL;
 	packed--;
 	min_pack_length++;
@@ -169,12 +174,8 @@ int mi_create(const char *name,uint keys,MI_KEYDEF *keydefs,
   /* We can't use checksum with static length rows */
   if (!(options & HA_OPTION_PACK_RECORD))
     options&= ~HA_OPTION_CHECKSUM;
-  if (options & (HA_OPTION_PACK_RECORD | HA_OPTION_COMPRESS_RECORD))
-    min_pack_length+=varchar_count;		/* Min length to pack */
-  else
-  {
-    min_pack_length+=varchar_length+2*varchar_count;
-  }
+  if (!(options & (HA_OPTION_PACK_RECORD | HA_OPTION_COMPRESS_RECORD)))
+    min_pack_length+= varchar_length;
   if (flags & HA_CREATE_TMP_TABLE)
   {
     options|= HA_OPTION_TMP_TABLE;
@@ -194,11 +195,10 @@ int mi_create(const char *name,uint keys,MI_KEYDEF *keydefs,
       test(test_all_bits(options, HA_OPTION_CHECKSUM | HA_PACK_RECORD));
   min_pack_length+=packed;
 
-  if (!ci->data_file_length)
+  if (!ci->data_file_length && ci->max_rows)
   {
-    if (ci->max_rows == 0 || pack_reclength == INT_MAX32)
-      ci->data_file_length= INT_MAX32-1;		/* Should be enough */
-    else if ((~(ulonglong) 0)/ci->max_rows < (ulonglong) pack_reclength)
+    if (pack_reclength == INT_MAX32 ||
+             (~(ulonglong) 0)/ci->max_rows < (ulonglong) pack_reclength)
       ci->data_file_length= ~(ulonglong) 0;
     else
       ci->data_file_length=(ulonglong) ci->max_rows*pack_reclength;
@@ -223,7 +223,7 @@ int mi_create(const char *name,uint keys,MI_KEYDEF *keydefs,
       reclength=pointer+1;		/* reserve place for delete link */
   }
   else
-    reclength+=long_varchar_count;	/* We need space for this! */
+    reclength+= long_varchar_count;	/* We need space for varchar! */
 
   max_key_length=0; tot_length=0 ; key_segs=0;
   fulltext_keys=0;
@@ -255,16 +255,19 @@ int mi_create(const char *name,uint keys,MI_KEYDEF *keydefs,
         /*
            called by myisamchk - i.e. table structure was taken from
            MYI file and SPATIAL key *does have* additional sp_segs keysegs.
-           We'd better delete them now
+           keydef->seg here points right at the GEOMETRY segment,
+           so we only need to decrease keydef->keysegs.
+           (see recreate_table() in mi_check.c)
         */
-        keydef->keysegs-=sp_segs;
+        keydef->keysegs-=sp_segs-1;
       }
 
       for (j=0, keyseg=keydef->seg ; (int) j < keydef->keysegs ;
 	   j++, keyseg++)
       {
         if (keyseg->type != HA_KEYTYPE_BINARY &&
-	    keyseg->type != HA_KEYTYPE_VARBINARY)
+	    keyseg->type != HA_KEYTYPE_VARBINARY1 &&
+            keyseg->type != HA_KEYTYPE_VARBINARY2)
         {
           my_errno=HA_WRONG_CREATE_OPTION;
           goto err;
@@ -279,8 +282,7 @@ int mi_create(const char *name,uint keys,MI_KEYDEF *keydefs,
       goto err;
 #endif /*HAVE_SPATIAL*/
     }
-    else
-    if (keydef->flag & HA_FULLTEXT)
+    else if (keydef->flag & HA_FULLTEXT)
     {
       keydef->flag=HA_FULLTEXT | HA_PACK_KEY | HA_VAR_LENGTH_KEY;
       options|=HA_OPTION_PACK_KEYS;             /* Using packed keys */
@@ -289,11 +291,22 @@ int mi_create(const char *name,uint keys,MI_KEYDEF *keydefs,
 	   j++, keyseg++)
       {
         if (keyseg->type != HA_KEYTYPE_TEXT &&
-	    keyseg->type != HA_KEYTYPE_VARTEXT)
+	    keyseg->type != HA_KEYTYPE_VARTEXT1 &&
+            keyseg->type != HA_KEYTYPE_VARTEXT2)
         {
           my_errno=HA_WRONG_CREATE_OPTION;
           goto err;
         }
+        if (!(keyseg->flag & HA_BLOB_PART) &&
+	    (keyseg->type == HA_KEYTYPE_VARTEXT1 ||
+             keyseg->type == HA_KEYTYPE_VARTEXT2))
+        {
+          /* Make a flag that this is a VARCHAR */
+          keyseg->flag|= HA_VAR_LENGTH_PART;
+          /* Store in bit_start number of bytes used to pack the length */
+          keyseg->bit_start= ((keyseg->type == HA_KEYTYPE_VARTEXT1)?
+                              1 : 2);
+        }
       }
 
       fulltext_keys++;
@@ -314,7 +327,7 @@ int mi_create(const char *name,uint keys,MI_KEYDEF *keydefs,
 
 	/* Only use HA_PACK_KEY when first segment is a variable length key */
 	if (!(keydef->seg[0].flag & (HA_SPACE_PACK | HA_BLOB_PART |
-				     HA_VAR_LENGTH)))
+				     HA_VAR_LENGTH_PART)))
 	{
 	  /* pack relative to previous key */
 	  keydef->flag&= ~HA_PACK_KEY;
@@ -348,12 +361,27 @@ int mi_create(const char *name,uint keys,MI_KEYDEF *keydefs,
 	case HA_KEYTYPE_UINT24:
 	case HA_KEYTYPE_INT8:
 	  keyseg->flag|= HA_SWAP_KEY;
-	  /* fall through */
+          break;
+        case HA_KEYTYPE_VARTEXT1:
+        case HA_KEYTYPE_VARTEXT2:
+        case HA_KEYTYPE_VARBINARY1:
+        case HA_KEYTYPE_VARBINARY2:
+          if (!(keyseg->flag & HA_BLOB_PART))
+          {
+            /* Make a flag that this is a VARCHAR */
+            keyseg->flag|= HA_VAR_LENGTH_PART;
+            /* Store in bit_start number of bytes used to pack the length */
+            keyseg->bit_start= ((keyseg->type == HA_KEYTYPE_VARTEXT1 ||
+                                 keyseg->type == HA_KEYTYPE_VARBINARY1) ?
+                                1 : 2);
+          }
+          break;
 	default:
 	  break;
 	}
 	if (keyseg->flag & HA_SPACE_PACK)
 	{
+          DBUG_ASSERT(!(keyseg->flag & HA_VAR_LENGTH_PART));
 	  keydef->flag |= HA_SPACE_PACK_USED | HA_VAR_LENGTH_KEY;
 	  options|=HA_OPTION_PACK_KEYS;		/* Using packed keys */
 	  length++;				/* At least one length byte */
@@ -364,8 +392,10 @@ int mi_create(const char *name,uint keys,MI_KEYDEF *keydefs,
 	    length+=2;
 	  }
 	}
-	if (keyseg->flag & (HA_VAR_LENGTH | HA_BLOB_PART))
+	if (keyseg->flag & (HA_VAR_LENGTH_PART | HA_BLOB_PART))
 	{
+          DBUG_ASSERT(!test_all_bits(keyseg->flag,
+                                    (HA_VAR_LENGTH_PART | HA_BLOB_PART)));
 	  keydef->flag|=HA_VAR_LENGTH_KEY;
 	  length++;				/* At least one length byte */
 	  options|=HA_OPTION_PACK_KEYS;		/* Using packed keys */
@@ -401,8 +431,8 @@ int mi_create(const char *name,uint keys,MI_KEYDEF *keydefs,
 	key_segs)
       share.state.rec_per_key_part[key_segs-1]=1L;
     length+=key_length;
-    keydef->block_length= MI_BLOCK_SIZE(length-real_length_diff,
-                                        pointer,MI_MAX_KEYPTR_SIZE);
+    keydef->block_length= (uint16) MI_BLOCK_SIZE(length-real_length_diff,
+                                                 pointer,MI_MAX_KEYPTR_SIZE);
     if (keydef->block_length > MI_MAX_KEY_BLOCK_LENGTH ||
         length >= MI_MAX_KEY_BUFF)
     {
@@ -446,6 +476,16 @@ int mi_create(const char *name,uint keys,MI_KEYDEF *keydefs,
 			       uniques * MI_UNIQUEDEF_SIZE +
 			       (key_segs + unique_key_parts)*HA_KEYSEG_SIZE+
 			       columns*MI_COLUMNDEF_SIZE);
+  DBUG_PRINT("info", ("info_length: %u", info_length));
+  /* There are only 16 bits for the total header length. */
+  if (info_length > 65535)
+  {
+    my_printf_error(0, "MyISAM table '%s' has too many columns and/or "
+                    "indexes and/or unique constraints.",
+                    MYF(0), name + dirname_length(name));
+    my_errno= HA_WRONG_CREATE_OPTION;
+    goto err;
+  }
 
   bmove(share.state.header.file_version,(byte*) myisam_file_magic,4);
   ci->old_options=options| (ci->old_options & HA_OPTION_TEMP_COMPRESS_RECORD ?
@@ -484,7 +524,7 @@ int mi_create(const char *name,uint keys,MI_KEYDEF *keydefs,
   mi_int2store(share.state.header.key_parts,key_segs);
   mi_int2store(share.state.header.unique_key_parts,unique_key_parts);
 
-  share.state.key_map = ((ulonglong) 1 << keys)-1;
+  mi_set_all_keys_active(share.state.key_map, keys);
   share.base.keystart = share.state.state.key_file_length=
     MY_ALIGN(info_length, myisam_block_size);
   share.base.max_key_block_length=max_key_block_length;
@@ -516,11 +556,24 @@ int mi_create(const char *name,uint keys,MI_KEYDEF *keydefs,
   if (! (flags & HA_DONT_TOUCH_DATA))
     share.state.create_time= (long) time((time_t*) 0);
 
+  pthread_mutex_lock(&THR_LOCK_myisam);
+
   if (ci->index_file_name)
   {
-    fn_format(filename, ci->index_file_name,"",MI_NAME_IEXT,4);
-    fn_format(linkname,name, "",MI_NAME_IEXT,4);
-    linkname_ptr=linkname;
+    if (options & HA_OPTION_TMP_TABLE)
+    {
+      char *path;
+      /* chop off the table name, tempory tables use generated name */
+      if ((path= strrchr(ci->index_file_name, FN_LIBCHAR)))
+        *path= '\0';
+      fn_format(filename, name, ci->index_file_name, MI_NAME_IEXT,
+                MY_REPLACE_DIR | MY_UNPACK_FILENAME);
+    }
+    else
+      fn_format(filename, ci->index_file_name, "",
+                MI_NAME_IEXT, MY_UNPACK_FILENAME);
+    fn_format(linkname, name, "", MI_NAME_IEXT, MY_UNPACK_FILENAME);
+    linkname_ptr= linkname;
     /*
       Don't create the table if the link or file exists to ensure that one
       doesn't accidently destroy another table.
@@ -533,7 +586,7 @@ int mi_create(const char *name,uint keys,MI_KEYDEF *keydefs,
 					     32 : 0));
     linkname_ptr=0;
     /* Replace the current file */
-    create_flag=MY_DELETE_OLD;
+    create_flag=(flags & HA_CREATE_KEEP_FILES) ? 0 : MY_DELETE_OLD;
   }
 
   /*
@@ -574,16 +627,27 @@ int mi_create(const char *name,uint keys,MI_KEYDEF *keydefs,
     {
       if (ci->data_file_name)
       {
-	fn_format(filename, ci->data_file_name,"",MI_NAME_DEXT,4);
-	fn_format(linkname, name, "",MI_NAME_DEXT,4);
-	linkname_ptr=linkname;
-	create_flag=0;
+        if (options & HA_OPTION_TMP_TABLE)
+        {
+          char *path;
+          /* chop off the table name, tempory tables use generated name */
+          if ((path= strrchr(ci->data_file_name, FN_LIBCHAR)))
+            *path= '\0';
+          fn_format(filename, name, ci->data_file_name, MI_NAME_DEXT,
+                    MY_REPLACE_DIR | MY_UNPACK_FILENAME);
+        }
+        else
+          fn_format(filename, ci->data_file_name, "",
+                    MI_NAME_DEXT, MY_UNPACK_FILENAME);
+        fn_format(linkname, name, "", MI_NAME_DEXT, MY_UNPACK_FILENAME);
+        linkname_ptr= linkname;
+        create_flag= 0;
       }
       else
       {
 	fn_format(filename,name,"",MI_NAME_DEXT,4);
 	linkname_ptr=0;
-	create_flag=MY_DELETE_OLD;
+        create_flag=(flags & HA_CREATE_KEEP_FILES) ? 0 : MY_DELETE_OLD;
       }
       if ((dfile=
 	   my_create_with_symlink(linkname_ptr, filename, 0, create_mode,
@@ -593,6 +657,7 @@ int mi_create(const char *name,uint keys,MI_KEYDEF *keydefs,
     errpos=3;
   }
 
+  DBUG_PRINT("info", ("write state info and base info"));
   if (mi_state_info_write(file, &share.state, 2) ||
       mi_base_info_write(file, &share.base))
     goto err;
@@ -606,6 +671,7 @@ int mi_create(const char *name,uint keys,MI_KEYDEF *keydefs,
 #endif
 
   /* Write key and keyseg definitions */
+  DBUG_PRINT("info", ("write key and keyseg definitions"));
   for (i=0 ; i < share.base.keys - uniques; i++)
   {
     uint sp_segs=(keydefs[i].flag & HA_SPATIAL) ? 2*SPDIMS : 0;
@@ -620,10 +686,12 @@ int mi_create(const char *name,uint keys,MI_KEYDEF *keydefs,
     {
       HA_KEYSEG sseg;
       sseg.type=SPTYPE;
-      sseg.language= 7;
+      sseg.language= 7;                         /* Binary */
       sseg.null_bit=0;
       sseg.bit_start=0;
       sseg.bit_end=0;
+      sseg.bit_length= 0;
+      sseg.bit_pos= 0;
       sseg.length=SPLEN;
       sseg.null_pos=0;
       sseg.start=j*SPLEN;
@@ -641,7 +709,7 @@ int mi_create(const char *name,uint keys,MI_KEYDEF *keydefs,
   {
     tmp_keydef.keysegs=1;
     tmp_keydef.flag=		HA_UNIQUE_CHECK;
-    tmp_keydef.block_length=	myisam_block_size;
+    tmp_keydef.block_length=	(uint16)myisam_block_size;
     tmp_keydef.keylength=	MI_UNIQUE_HASH_LENGTH + pointer;
     tmp_keydef.minlength=tmp_keydef.maxlength=tmp_keydef.keylength;
     tmp_keyseg.type=		MI_UNIQUE_HASH_TYPE;
@@ -654,16 +722,38 @@ int mi_create(const char *name,uint keys,MI_KEYDEF *keydefs,
   }
 
   /* Save unique definition */
+  DBUG_PRINT("info", ("write unique definitions"));
   for (i=0 ; i < share.state.header.uniques ; i++)
   {
+    HA_KEYSEG *keyseg_end;
+    keyseg= uniquedefs[i].seg;
     if (mi_uniquedef_write(file, &uniquedefs[i]))
       goto err;
-    for (j=0 ; j < uniquedefs[i].keysegs ; j++)
+    for (keyseg= uniquedefs[i].seg, keyseg_end= keyseg+ uniquedefs[i].keysegs;
+         keyseg < keyseg_end;
+         keyseg++)
     {
-      if (mi_keyseg_write(file, &uniquedefs[i].seg[j]))
+      switch (keyseg->type) {
+      case HA_KEYTYPE_VARTEXT1:
+      case HA_KEYTYPE_VARTEXT2:
+      case HA_KEYTYPE_VARBINARY1:
+      case HA_KEYTYPE_VARBINARY2:
+        if (!(keyseg->flag & HA_BLOB_PART))
+        {
+          keyseg->flag|= HA_VAR_LENGTH_PART;
+          keyseg->bit_start= ((keyseg->type == HA_KEYTYPE_VARTEXT1 ||
+                               keyseg->type == HA_KEYTYPE_VARBINARY1) ?
+                              1 : 2);
+        }
+        break;
+      default:
+        break;
+      }
+      if (mi_keyseg_write(file, keyseg))
 	goto err;
     }
   }
+  DBUG_PRINT("info", ("write field definitions"));
   for (i=0 ; i < share.base.fields ; i++)
     if (mi_recinfo_write(file, &recinfo[i]))
       goto err;
@@ -678,6 +768,7 @@ int mi_create(const char *name,uint keys,MI_KEYDEF *keydefs,
 #endif
 
 	/* Enlarge files */
+  DBUG_PRINT("info", ("enlarge to keystart: %lu", (ulong) share.base.keystart));
   if (my_chsize(file,(ulong) share.base.keystart,0,MYF(0)))
     goto err;
 
@@ -724,19 +815,23 @@ err:
 
 uint mi_get_pointer_length(ulonglong file_length, uint def)
 {
+  DBUG_ASSERT(def >= 2 && def <= 7);
   if (file_length)				/* If not default */
   {
-    if (file_length >= (longlong) 1 << 56)
+#ifdef NOT_YET_READY_FOR_8_BYTE_POINTERS
+    if (file_length >= ULL(1) << 56)
       def=8;
-    if (file_length >= (longlong) 1 << 48)
+    else
+#endif
+    if (file_length >= ULL(1) << 48)
       def=7;
-    if (file_length >= (longlong) 1 << 40)
+    else if (file_length >= ULL(1) << 40)
       def=6;
-    else if (file_length >= (longlong) 1 << 32)
+    else if (file_length >= ULL(1) << 32)
       def=5;
-    else if (file_length >= (1L << 24))
+    else if (file_length >= ULL(1) << 24)
       def=4;
-    else if (file_length >= (1L << 16))
+    else if (file_length >= ULL(1) << 16)
       def=3;
     else
       def=2;
