@@ -1,9 +1,8 @@
-/* Copyright (C) 2000 MySQL AB & MySQL Finland AB & TCX DataKonsult AB
+/* Copyright (C) 2000-2006 MySQL AB
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
-   the Free Software Foundation; either version 2 of the License, or
-   (at your option) any later version.
+   the Free Software Foundation; version 2 of the License.
 
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
@@ -139,8 +138,9 @@ int _create_index_by_sort(MI_SORT_PARAM *info,my_bool no_messages,
 
   while (memavl >= MIN_SORT_MEMORY)
   {
-    if ((my_off_t) (records+1)*(sort_length+sizeof(char*)) <=
-	(my_off_t) memavl)
+    if ((records < UINT_MAX32) && 
+       ((my_off_t) (records + 1) * 
+        (sort_length + sizeof(char*)) <= (my_off_t) memavl))
       keys= records+1;
     else
       do
@@ -152,7 +152,7 @@ int _create_index_by_sort(MI_SORT_PARAM *info,my_bool no_messages,
             keys < (uint) maxbuffer)
 	{
 	  mi_check_print_error(info->sort_info->param,
-			       "sort_buffer_size is to small");
+			       "myisam_sort_buffer_size is too small");
 	  goto err;
 	}
       }
@@ -176,7 +176,7 @@ int _create_index_by_sort(MI_SORT_PARAM *info,my_bool no_messages,
   }
   if (memavl < MIN_SORT_MEMORY)
   {
-    mi_check_print_error(info->sort_info->param,"Sort buffer to small"); /* purecov: tested */
+    mi_check_print_error(info->sort_info->param,"MyISAM sort buffer too small"); /* purecov: tested */
     goto err; /* purecov: tested */
   }
   (*info->lock_in_memory)(info->sort_info->param);/* Everything is allocated */
@@ -221,9 +221,9 @@ int _create_index_by_sort(MI_SORT_PARAM *info,my_bool no_messages,
 
   if (my_b_inited(&tempfile_for_exceptions))
   {
-    MI_INFO *index=info->sort_info->info;
+    MI_INFO *idx=info->sort_info->info;
     uint     keyno=info->key;
-    uint     key_length, ref_length=index->s->rec_reflength;
+    uint     key_length, ref_length=idx->s->rec_reflength;
 
     if (!no_messages)
       printf("  - Adding exceptions\n"); /* purecov: tested */
@@ -236,7 +236,7 @@ int _create_index_by_sort(MI_SORT_PARAM *info,my_bool no_messages,
         && !my_b_read(&tempfile_for_exceptions,(byte*)sort_keys,
 		      (uint) key_length))
     {
-	if (_mi_ck_write(index,keyno,(uchar*) sort_keys,key_length-ref_length))
+	if (_mi_ck_write(idx,keyno,(uchar*) sort_keys,key_length-ref_length))
 	  goto err;
     }
   }
@@ -370,7 +370,7 @@ pthread_handler_t thr_find_all_keys(void *arg)
               keys < (uint) maxbuffer)
           {
             mi_check_print_error(sort_param->sort_info->param,
-                                 "sort_buffer_size is to small");
+                                 "myisam_sort_buffer_size is too small");
             goto err;
           }
         }
@@ -398,7 +398,7 @@ pthread_handler_t thr_find_all_keys(void *arg)
     if (memavl < MIN_SORT_MEMORY)
     {
       mi_check_print_error(sort_param->sort_info->param,
-                           "Sort buffer too small");
+                           "MyISAM sort buffer too small");
       goto err; /* purecov: tested */
     }
 
@@ -774,7 +774,7 @@ static int NEAR_F merge_many_buff(MI_SORT_PARAM *info, uint keys,
     {
       if (merge_buffers(info,keys,from_file,to_file,sort_keys,lastbuff++,
                         buffpek+i,buffpek+i+MERGEBUFF-1))
-        break; /* purecov: inspected */
+        goto cleanup;
     }
     if (merge_buffers(info,keys,from_file,to_file,sort_keys,lastbuff++,
                       buffpek+i,buffpek+ *maxbuffer))
@@ -784,6 +784,7 @@ static int NEAR_F merge_many_buff(MI_SORT_PARAM *info, uint keys,
     temp=from_file; from_file=to_file; to_file=temp;
     *maxbuffer= (int) (lastbuff-buffpek)-1;
   }
+cleanup:
   close_cached_file(to_file);                   /* This holds old result */
   if (to_file == t_file)
     *t_file=t_file2;                            /* Copy result file */
