@@ -2,8 +2,7 @@
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
-   the Free Software Foundation; either version 2 of the License, or
-   (at your option) any later version.
+   the Free Software Foundation; version 2 of the License.
 
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
@@ -17,6 +16,8 @@
 #ifndef MGMAPI_H
 #define MGMAPI_H
 
+#define NDB_MGM_MAX_LOGLEVEL 15
+
 /**
  * @mainpage MySQL Cluster Management API
  *
@@ -541,6 +542,16 @@ extern "C" {
    */
   const char *ndb_mgm_get_connectstring(NdbMgmHandle handle, char *buf, int buf_sz);
 
+  /**
+   * Sets the number of seconds to wait for connect(2) during ndb_mgm_connect
+   * Default is no timeout
+   *
+   * @param handle  NdbMgmHandle
+   * @param seconds number of seconds
+   * @return non-zero on success
+   */
+  int ndb_mgm_set_connect_timeout(NdbMgmHandle handle, unsigned int seconds);
+
   /**
    * Connects to a management server. Connectstring is set by
    * ndb_mgm_set_connectstring().
