@@ -2,8 +2,7 @@
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
-   the Free Software Foundation; either version 2 of the License, or
-   (at your option) any later version.
+   the Free Software Foundation; version 2 of the License.
 
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
@@ -24,6 +23,7 @@ class SocketClient
 {
   NDB_SOCKET_TYPE m_sockfd;
   struct sockaddr_in m_servaddr;
+  unsigned int m_connect_timeout_sec;
   unsigned short m_port;
   char *m_server_name;
   SocketAuthenticator *m_auth;
@@ -31,7 +31,17 @@ public:
   SocketClient(const char *server_name, unsigned short port, SocketAuthenticator *sa = 0);
   ~SocketClient();
   bool init();
-  NDB_SOCKET_TYPE connect();
+  void set_port(unsigned short port) {
+    m_port = port;
+    m_servaddr.sin_port = htons(m_port);
+  };
+  void set_connect_timeout(unsigned int s) {
+    m_connect_timeout_sec= s;
+  }
+  unsigned short get_port() { return m_port; };
+  char *get_server_name() { return m_server_name; };
+  int bind(const char* toaddress, unsigned short toport);
+  NDB_SOCKET_TYPE connect(const char* toaddress = 0, unsigned short port = 0);
   bool close();
 };
 
