@@ -2,8 +2,7 @@
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
-   the Free Software Foundation; either version 2 of the License, or
-   (at your option) any later version.
+   the Free Software Foundation; version 2 of the License.
 
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
@@ -78,11 +77,7 @@ Cmvmi::Cmvmi(const Configuration & conf) :
   addRecSignal(GSN_OPEN_COMREQ,  &Cmvmi::execOPEN_COMREQ);
   addRecSignal(GSN_TEST_ORD,  &Cmvmi::execTEST_ORD);
 
-  addRecSignal(GSN_STATISTICS_REQ,  &Cmvmi::execSTATISTICS_REQ);
   addRecSignal(GSN_TAMPER_ORD,  &Cmvmi::execTAMPER_ORD);
-  addRecSignal(GSN_SET_VAR_REQ,  &Cmvmi::execSET_VAR_REQ);
-  addRecSignal(GSN_SET_VAR_CONF,  &Cmvmi::execSET_VAR_CONF);
-  addRecSignal(GSN_SET_VAR_REF,  &Cmvmi::execSET_VAR_REF);
   addRecSignal(GSN_STOP_ORD,  &Cmvmi::execSTOP_ORD);
   addRecSignal(GSN_START_ORD,  &Cmvmi::execSTART_ORD);
   addRecSignal(GSN_EVENT_SUBSCRIBE_REQ, 
@@ -91,6 +86,7 @@ Cmvmi::Cmvmi(const Configuration & conf) :
   addRecSignal(GSN_DUMP_STATE_ORD, &Cmvmi::execDUMP_STATE_ORD);
 
   addRecSignal(GSN_TESTSIG, &Cmvmi::execTESTSIG);
+  addRecSignal(GSN_NODE_START_REP, &Cmvmi::execNODE_START_REP, true);
   
   subscriberPool.setSize(5);
 
@@ -136,6 +132,7 @@ Cmvmi::~Cmvmi()
 
 #ifdef ERROR_INSERT
 NodeBitmask c_error_9000_nodes_mask;
+extern Uint32 MAX_RECEIVED_SIGNALS;
 #endif
 
 void Cmvmi::execNDB_TAMPER(Signal* signal) 
@@ -165,6 +162,22 @@ void Cmvmi::execNDB_TAMPER(Signal* signal)
     kill(getpid(), SIGABRT);
   }
 #endif
+
+#ifdef ERROR_INSERT
+  if (signal->theData[0] == 9003)
+  {
+    if (MAX_RECEIVED_SIGNALS < 1024)
+    {
+      MAX_RECEIVED_SIGNALS = 1024;
+    }
+    else
+    {
+      MAX_RECEIVED_SIGNALS = 1 + (rand() % 128);
+    }
+    ndbout_c("MAX_RECEIVED_SIGNALS: %d", MAX_RECEIVED_SIGNALS);
+    CLEAR_ERROR_INSERT_VALUE;
+  }
+#endif
 }//execNDB_TAMPER()
 
 void Cmvmi::execSET_LOGLEVELORD(Signal* signal) 
@@ -234,6 +247,7 @@ void Cmvmi::execEVENT_REP(Signal* signal)
 void
 Cmvmi::execEVENT_SUBSCRIBE_REQ(Signal * signal){
   EventSubscribeReq * subReq = (EventSubscribeReq *)&signal->theData[0];
+  Uint32 senderRef = signal->getSendersBlockRef();
   SubscriberPtr ptr;
   jamEntry();
   DBUG_ENTER("Cmvmi::execEVENT_SUBSCRIBE_REQ");
@@ -251,7 +265,7 @@ Cmvmi::execEVENT_SUBSCRIBE_REQ(Signal * signal){
      * Create a new one
      */
     if(subscribers.seize(ptr) == false){
-      sendSignal(subReq->blockRef, GSN_EVENT_SUBSCRIBE_REF, signal, 1, JBB);
+      sendSignal(senderRef, GSN_EVENT_SUBSCRIBE_REF, signal, 1, JBB);
       return;
     }
     ptr.p->logLevel.clear();
@@ -278,7 +292,7 @@ Cmvmi::execEVENT_SUBSCRIBE_REQ(Signal * signal){
   }
   
   signal->theData[0] = ptr.i;
-  sendSignal(ptr.p->blockRef, GSN_EVENT_SUBSCRIBE_CONF, signal, 1, JBB);
+  sendSignal(senderRef, GSN_EVENT_SUBSCRIBE_CONF, signal, 1, JBB);
   DBUG_VOID_RETURN;
 }
 
@@ -339,6 +353,16 @@ void Cmvmi::execSTTOR(Signal* signal)
   jamEntry();
   if (theStartPhase == 1){
     jam();
+
+    if(theConfig.lockPagesInMainMemory() == 1)
+    {
+      int res = NdbMem_MemLockAll(0);
+      if(res != 0){
+	g_eventLogger.warning("Failed to memlock pages");
+	warningEvent("Failed to memlock pages");
+      }
+    }
+    
     sendSTTORRY(signal);
     return;
   } else if (theStartPhase == 3) {
@@ -358,18 +382,6 @@ void Cmvmi::execSTTOR(Signal* signal)
     signal->theData[2] = NodeInfo::REP;
     execOPEN_COMREQ(signal);    
     globalData.theStartLevel = NodeState::SL_STARTED;
-    sendSTTORRY(signal);
-  } else {
-    jam();
-
-    if(theConfig.lockPagesInMainMemory()){
-      int res = NdbMem_MemLockAll();
-      if(res != 0){
-	g_eventLogger.warning("Failed to memlock pages");
-	warningEvent("Failed to memlock pages");
-      }
-    }
-    
     sendSTTORRY(signal);
   }
 }
@@ -424,7 +436,8 @@ void Cmvmi::execOPEN_COMREQ(Signal* signal)
   if(len == 2){
 
 #ifdef ERROR_INSERT
-    if (! (ERROR_INSERTED(9000) && c_error_9000_nodes_mask.get(tStartingNode)))
+    if (! ((ERROR_INSERTED(9000) || ERROR_INSERTED(9002)) 
+	   && c_error_9000_nodes_mask.get(tStartingNode)))
 #endif
     {
       globalTransporterRegistry.do_connect(tStartingNode);
@@ -445,7 +458,8 @@ void Cmvmi::execOPEN_COMREQ(Signal* signal)
 	jam();
 
 #ifdef ERROR_INSERT
-	if (ERROR_INSERTED(9000) && c_error_9000_nodes_mask.get(i))
+	if ((ERROR_INSERTED(9000) || ERROR_INSERTED(9002))
+	    && c_error_9000_nodes_mask.get(i))
 	  continue;
 #endif
 	
@@ -696,29 +710,12 @@ Cmvmi::execTEST_ORD(Signal * signal){
       // Do nothing
       break;
     }
+    globalSignalLoggers.flushSignalLog();
   }
 
 #endif
 }
 
-void Cmvmi::execSTATISTICS_REQ(Signal* signal) 
-{
-  // TODO Note ! This is only a test implementation...
-
-  static int stat1 = 0;
-  jamEntry();
-
-  //ndbout << "data 1: " << signal->theData[1];
-
-  int x = signal->theData[0];
-  stat1++;
-  signal->theData[0] = stat1;
-  sendSignal(x, GSN_STATISTICS_CONF, signal, 7, JBB);
-
-}//execSTATISTICS_REQ()
-
-
-
 void Cmvmi::execSTOP_ORD(Signal* signal) 
 {
   jamEntry();
@@ -786,6 +783,21 @@ Cmvmi::execSTART_ORD(Signal* signal) {
   
   if(globalData.theStartLevel == NodeState::SL_CMVMI){
     jam();
+
+    if(theConfig.lockPagesInMainMemory() == 2)
+    {
+      int res = NdbMem_MemLockAll(1);
+      if(res != 0)
+      {
+	g_eventLogger.warning("Failed to memlock pages");
+	warningEvent("Failed to memlock pages");
+      }
+      else
+      {
+	g_eventLogger.info("Locked future allocations");
+      }
+    }
+    
     globalData.theStartLevel  = NodeState::SL_STARTING;
     globalData.theRestartFlag = system_started;
     /**
@@ -822,7 +834,7 @@ void Cmvmi::execTAMPER_ORD(Signal* signal)
   // to be able to indicate if we really introduced an error.
 #ifdef ERROR_INSERT
   TamperOrd* const tamperOrd = (TamperOrd*)&signal->theData[0];
-  
+  signal->theData[2] = 0;
   signal->theData[1] = tamperOrd->errorNo;
   signal->theData[0] = 5;
   sendSignal(DBDIH_REF, GSN_DIHNDBTAMPER, signal, 3,JBB);
@@ -830,160 +842,6 @@ void Cmvmi::execTAMPER_ORD(Signal* signal)
 
 }//execTAMPER_ORD()
 
-
-
-void Cmvmi::execSET_VAR_REQ(Signal* signal) 
-{
-#if 0
-
-  SetVarReq* const setVarReq = (SetVarReq*)&signal->theData[0];
-  ConfigParamId var = setVarReq->variable();
-  jamEntry();
-  switch (var) {
-    
-    // NDBCNTR_REF
-    
-    // DBTC
-  case TransactionDeadlockDetectionTimeout:
-  case TransactionInactiveTime:
-  case NoOfConcurrentProcessesHandleTakeover:
-    sendSignal(DBTC_REF, GSN_SET_VAR_REQ, signal, 3, JBB);
-    break;
-    
-    // DBDIH
-  case TimeBetweenLocalCheckpoints:
-  case TimeBetweenGlobalCheckpoints:
-    sendSignal(DBDIH_REF, GSN_SET_VAR_REQ, signal, 3, JBB);
-    break;
-
-    // DBLQH
-  case NoOfConcurrentCheckpointsDuringRestart:
-  case NoOfConcurrentCheckpointsAfterRestart:
-    sendSignal(DBLQH_REF, GSN_SET_VAR_REQ, signal, 3, JBB);
-    break;
-
-    // DBACC
-  case NoOfDiskPagesToDiskDuringRestartACC:
-  case NoOfDiskPagesToDiskAfterRestartACC:
-    sendSignal(DBACC_REF, GSN_SET_VAR_REQ, signal, 3, JBB);
-    break;
-
-    // DBTUP
-  case NoOfDiskPagesToDiskDuringRestartTUP:
-  case NoOfDiskPagesToDiskAfterRestartTUP:
-    sendSignal(DBTUP_REF, GSN_SET_VAR_REQ, signal, 3, JBB);
-    break;
-
-    // DBDICT
-
-    // NDBCNTR
-  case TimeToWaitAlive:
-
-    // QMGR
-  case HeartbeatIntervalDbDb: // TODO ev till Ndbcnt ocks
-  case HeartbeatIntervalDbApi:
-  case ArbitTimeout:
-    sendSignal(QMGR_REF, GSN_SET_VAR_REQ, signal, 3, JBB);
-    break;
-
-    // NDBFS
-
-    // CMVMI
-  case MaxNoOfSavedMessages:
-  case LockPagesInMainMemory:
-  case TimeBetweenWatchDogCheck:
-  case StopOnError:
-    handleSET_VAR_REQ(signal);
-    break;
-
-
-    // Not possible to update (this could of course be handled by each block
-    // instead but I havn't investigated where they belong)
-  case Id:
-  case ExecuteOnComputer:
-  case ShmKey:
-  case MaxNoOfConcurrentOperations:
-  case MaxNoOfConcurrentTransactions:
-  case MemorySpaceIndexes:
-  case MemorySpaceTuples:
-  case MemoryDiskPages:
-  case NoOfFreeDiskClusters:
-  case NoOfDiskClusters:
-  case NoOfFragmentLogFiles:
-  case NoOfDiskClustersPerDiskFile:
-  case NoOfDiskFiles:
-  case MaxNoOfSavedEvents:
-  default:
-
-    int mgmtSrvr = setVarReq->mgmtSrvrBlockRef();
-    sendSignal(mgmtSrvr, GSN_SET_VAR_REF, signal, 0, JBB);
-  } // switch
-
-#endif
-}//execSET_VAR_REQ()
-
-
-void Cmvmi::execSET_VAR_CONF(Signal* signal) 
-{
-  int mgmtSrvr = signal->theData[0];
-  sendSignal(mgmtSrvr, GSN_SET_VAR_CONF, signal, 0, JBB);
-
-}//execSET_VAR_CONF()
-
-
-void Cmvmi::execSET_VAR_REF(Signal* signal) 
-{
-  int mgmtSrvr = signal->theData[0];
-  sendSignal(mgmtSrvr, GSN_SET_VAR_REF, signal, 0, JBB);
-
-}//execSET_VAR_REF()
-
-
-void Cmvmi::handleSET_VAR_REQ(Signal* signal) {
-#if 0
-  SetVarReq* const setVarReq = (SetVarReq*)&signal->theData[0];
-  ConfigParamId var = setVarReq->variable();
-  int val = setVarReq->value();
-
-  switch (var) {
-  case MaxNoOfSavedMessages:
-    theConfig.maxNoOfErrorLogs(val);
-    sendSignal(CMVMI_REF, GSN_SET_VAR_CONF, signal, 1, JBB);
-    break;
-    
-  case LockPagesInMainMemory:
-    int result;
-    if (val == 0) {
-      result = NdbMem_MemUnlockAll();
-    }
-    else {
-      result = NdbMem_MemLockAll();
-    }
-    if (result == 0) {
-      sendSignal(CMVMI_REF, GSN_SET_VAR_CONF, signal, 1, JBB);
-    }
-    else {
-      sendSignal(CMVMI_REF, GSN_SET_VAR_REF, signal, 1, JBB);
-    }
-    break;
-
-  case TimeBetweenWatchDogCheck:
-    theConfig.timeBetweenWatchDogCheck(val);
-    sendSignal(CMVMI_REF, GSN_SET_VAR_CONF, signal, 1, JBB);
-    break;
-
-  case StopOnError:
-    theConfig.stopOnError(val);
-    sendSignal(CMVMI_REF, GSN_SET_VAR_CONF, signal, 1, JBB);
-    break;
-    
-  default:
-    sendSignal(CMVMI_REF, GSN_SET_VAR_REF, signal, 1, JBB);
-    return;
-  } // switch
-#endif
-}
-
 #ifdef VM_TRACE
 class RefSignalTest {
 public:
@@ -1088,6 +946,24 @@ Cmvmi::execDUMP_STATE_ORD(Signal* signal)
     }
   }
   
+  if (arg == DumpStateOrd::CmvmiDumpSubscriptions)
+  {
+    SubscriberPtr ptr;
+    subscribers.first(ptr);  
+    g_eventLogger.info("List subscriptions:");
+    while(ptr.i != RNIL)
+    {
+      g_eventLogger.info("Subscription: %u, nodeId: %u, ref: 0x%x",
+                         ptr.i,  refToNode(ptr.p->blockRef), ptr.p->blockRef);
+      for(Uint32 i = 0; i < LogLevel::LOGLEVEL_CATEGORIES; i++)
+      {
+        Uint32 level = ptr.p->logLevel.getLogLevel((LogLevel::EventCategory)i);
+        g_eventLogger.info("Category %u Level %u", i, level);
+      }
+      subscribers.next(ptr);
+    }
+  }
+
   if (arg == DumpStateOrd::CmvmiDumpLongSignalMemory){
     infoEvent("Cmvmi: g_sectionSegmentPool size: %d free: %d",
 	      g_sectionSegmentPool.getSize(),
@@ -1143,9 +1019,9 @@ Cmvmi::execDUMP_STATE_ORD(Signal* signal)
   }
 
 #ifdef ERROR_INSERT
-  if (arg == 9000)
+  if (arg == 9000 || arg == 9002)
   {
-    SET_ERROR_INSERT_VALUE(9000);
+    SET_ERROR_INSERT_VALUE(arg);
     for (Uint32 i = 1; i<signal->getLength(); i++)
       c_error_9000_nodes_mask.set(signal->theData[i]);
   }
@@ -1192,6 +1068,17 @@ Cmvmi::execDUMP_STATE_ORD(Signal* signal)
 #endif
 }//Cmvmi::execDUMP_STATE_ORD()
 
+void
+Cmvmi::execNODE_START_REP(Signal* signal)
+{
+#ifdef ERROR_INSERT
+  if (ERROR_INSERTED(9002) && signal->theData[0] == getOwnNodeId())
+  {
+    signal->theData[0] = 9001;
+    execDUMP_STATE_ORD(signal);
+  }
+#endif
+}
 
 BLOCK_FUNCTIONS(Cmvmi)
 
@@ -1233,7 +1120,7 @@ Cmvmi::execTESTSIG(Signal* signal){
     fprintf(stdout, "\n");
     
     for(i = 0; i<signal->header.m_noOfSections; i++){
-      SegmentedSectionPtr ptr;
+      SegmentedSectionPtr ptr(0,0,0);
       ndbout_c("-- Section %d --", i);
       signal->getSection(ptr, i);
       ndbrequire(ptr.p != 0);
@@ -1291,7 +1178,7 @@ Cmvmi::execTESTSIG(Signal* signal){
     LinearSectionPtr ptr[3];
     const Uint32 secs = signal->getNoOfSections();
     for(i = 0; i<secs; i++){
-      SegmentedSectionPtr sptr;
+      SegmentedSectionPtr sptr(0,0,0);
       signal->getSection(sptr, i);
       ptr[i].sz = sptr.sz;
       ptr[i].p = new Uint32[sptr.sz];
@@ -1340,7 +1227,7 @@ Cmvmi::execTESTSIG(Signal* signal){
     LinearSectionPtr ptr[3];
     const Uint32 secs = signal->getNoOfSections();
     for(i = 0; i<secs; i++){
-      SegmentedSectionPtr sptr;
+      SegmentedSectionPtr sptr(0,0,0);
       signal->getSection(sptr, i);
       ptr[i].sz = sptr.sz;
       ptr[i].p = new Uint32[sptr.sz];
@@ -1406,7 +1293,7 @@ Cmvmi::execTESTSIG(Signal* signal){
     const Uint32 secs = signal->getNoOfSections();
     memset(g_test, 0, sizeof(g_test));
     for(i = 0; i<secs; i++){
-      SegmentedSectionPtr sptr;
+      SegmentedSectionPtr sptr(0,0,0);
       signal->getSection(sptr, i);
       g_test[i].sz = sptr.sz;
       g_test[i].p = new Uint32[sptr.sz];
