@@ -2,8 +2,7 @@
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
-   the Free Software Foundation; either version 2 of the License, or
-   (at your option) any later version.
+   the Free Software Foundation; version 2 of the License.
 
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
@@ -139,8 +138,9 @@
 
 #define ZNOT_FOUND 626
 #define ZALREADYEXIST 630
-#define ZINCONSISTENTHASHINDEX 892
 #define ZNOTUNIQUE 893
+
+#define ZINVALID_KEY 290
 #endif
 
 class Dbtc: public SimulatedBlock {
@@ -211,14 +211,6 @@ public:
     LTS_ACTIVE = 1
   };
 
-  enum TakeOverState {
-    TOS_NOT_DEFINED = 0,
-    TOS_IDLE = 1,
-    TOS_ACTIVE = 2,
-    TOS_COMPLETED = 3,
-    TOS_NODE_FAILED = 4
-  };
-
   enum FailState {
     FS_IDLE = 0,
     FS_LISTENING = 1,
@@ -310,6 +302,7 @@ public:
   /* WHEN THE TRIGGER IS DEACTIVATED.         */
   /* **************************************** */
   struct TcDefinedTriggerData {
+    TcDefinedTriggerData() {}
     /**
      * Trigger id, used to identify the trigger
      */
@@ -561,7 +554,7 @@ public:
     Uint32 expectedTransIdAI;
     AttributeBuffer transIdAI; // For accumulating TransId_AI
     
-    TcIndxReq tcIndxReq;
+    TcKeyReq tcIndxReq;
     UintR connectionIndex;
     UintR indexReadTcConnect; //
     
@@ -585,34 +578,8 @@ public:
    */
   ArrayPool<TcIndexOperation> c_theIndexOperationPool;
 
-  /**
-   * The list of index operations
-   */  
-  ArrayList<TcIndexOperation> c_theIndexOperations;
-
   UintR c_maxNumberOfIndexOperations;   
 
-  struct TcSeizedIndexOperation {
-    /**
-     * Next ptr (used in pool/list)
-     */
-    union {
-      Uint32 nextPool;
-      Uint32 nextList;
-    };
-    /**
-     * Prev pointer (used in list)
-     */
-    Uint32 prevList;
-  };
-
-  /**
-   * Pool of seized index operations
-   */
-  ArrayPool<TcSeizedIndexOperation> c_theSeizedIndexOperationPool;
-
-  typedef Ptr<TcSeizedIndexOperation> TcSeizedIndexOperationPtr;
-
   /************************** API CONNECT RECORD ***********************
    * The API connect record contains the connection record to which the
    * application connects.  
@@ -650,7 +617,7 @@ public:
 
   struct ApiConnectRecord {
     ApiConnectRecord(ArrayPool<TcFiredTriggerData> & firedTriggerPool,
-		     ArrayPool<TcSeizedIndexOperation> & seizedIndexOpPool):
+		     ArrayPool<TcIndexOperation> & seizedIndexOpPool):
       theFiredTriggers(firedTriggerPool),
       isIndexOp(false),
       theSeizedIndexOperations(seizedIndexOpPool) 
@@ -662,6 +629,7 @@ public:
     ConnectionState apiConnectstate;
     UintR transid[2];
     UintR firstTcConnect;
+    NdbNodeBitmask m_transaction_nodes; 
     
     //---------------------------------------------------
     // Second 16 byte cache line. Hot variables.
@@ -735,6 +703,7 @@ public:
     Uint8 tckeyrec; // ndrad frn R
     Uint8 tcindxrec;
     Uint8 apiFailState; // ndrad frn R
+    Uint8 singleUserMode;
     ReturnSignal returnsignal;
     Uint8 timeOutCounter;
     
@@ -751,7 +720,7 @@ public:
     
     // Index data
     
-    bool isIndexOp;      // Used to mark on-going TcKeyReq as indx table access
+    Uint8 isIndexOp;      // Used to mark on-going TcKeyReq as indx table access
     bool indexOpReturn;
     UintR noIndexOp;     // No outstanding index ops
 
@@ -763,7 +732,7 @@ public:
     UintR accumulatingIndexOp;
     UintR executingIndexOp;
     UintR tcIndxSendArray[6];
-    ArrayList<TcSeizedIndexOperation> theSeizedIndexOperations;
+    ArrayList<TcIndexOperation> theSeizedIndexOperations;
   };
   
   typedef Ptr<ApiConnectRecord> ApiConnectRecordPtr;
@@ -839,7 +808,7 @@ public:
     UintR savedState[LqhKeyConf::SignalLength];
     
     // Index data
-    bool isIndexOp; // Used to mark on-going TcKeyReq as index table access
+    Uint8 isIndexOp; // Used to mark on-going TcKeyReq as index table access
     UintR indexOp;
     UintR currentIndexId;
     UintR attrInfoLen;
@@ -897,11 +866,11 @@ public:
     UintR  hashValue;    /* THE HASH VALUE USED TO LOCATE FRAGMENT       */
     
     Uint8  distributionKeyIndicator;
-    Uint8  distributionGroupIndicator;
-    Uint8  distributionGroupType;
+    Uint8  m_special_hash; // collation or distribution key
+    Uint8  unused2;
     Uint8  lenAiInTckeyreq;  /* LENGTH OF ATTRIBUTE INFORMATION IN TCKEYREQ */
 
-    Uint8  distributionKey;
+    Uint8  fragmentDistributionKey;  /* DIH generation no */
 
     /**
      * EXECUTION MODE OF OPERATION                    
@@ -925,16 +894,16 @@ public:
     // Second 16 byte cache line in second 64 byte cache
     // line. Diverse use.
     //---------------------------------------------------
-    UintR  distributionGroup;
+    UintR  distributionKey;
     UintR  nextCacheRec;
-    UintR  distributionKeySize;
+    UintR  unused3;
     Uint32 scanInfo;
     
     //---------------------------------------------------
     // Third 16 byte cache line in second 64
     // byte cache line. Diverse use.
     //---------------------------------------------------
-    Uint32 scanNode;
+    Uint32 unused4;
     Uint32 scanTakeOverInd;
     UintR  firstKeybuf;   /* POINTER THE LINKED LIST OF KEY BUFFERS       */
     UintR  lastKeybuf;    /* VARIABLE POINTING TO THE LAST KEY BUFFER     */
@@ -958,7 +927,6 @@ public:
   struct HostRecord {
     HostState hostStatus;
     LqhTransState lqhTransStatus;
-    TakeOverState takeOverStatus;
     bool  inPackedList;
     UintR noOfPackedWordsLqh;
     UintR packedWordsLqh[26];
@@ -967,28 +935,58 @@ public:
     UintR noOfWordsTCINDXCONF;
     UintR packedWordsTCINDXCONF[30];
     BlockReference hostLqhBlockRef;
+
+    enum NodeFailBits
+    {
+      NF_TAKEOVER          = 0x1,
+      NF_CHECK_SCAN        = 0x2,
+      NF_CHECK_TRANSACTION = 0x4,
+      NF_CHECK_DROP_TAB    = 0x8,
+      NF_NODE_FAIL_BITS    = 0xF, // All bits...
+      NF_STARTED           = 0x10
+    };
+    Uint32 m_nf_bits;
+    NdbNodeBitmask m_lqh_trans_conf;
   }; /* p2c: size = 128 bytes */
   
   typedef Ptr<HostRecord> HostRecordPtr;
   
   /* *********** TABLE RECORD ********************************************* */
+
   /********************************************************/
   /* THIS RECORD CONTAINS THE CURRENT SCHEMA VERSION OF   */
   /* ALL TABLES IN THE SYSTEM.                            */
   /********************************************************/
   struct TableRecord {
+    TableRecord() {}
     Uint32 currentSchemaVersion;
-    Uint8 enabled;
-    Uint8 dropping;
+    Uint16 m_flags;
     Uint8 tableType;
-    Uint8 storedTable;
+    Uint8 singleUserMode;
+
+    enum {
+      TR_ENABLED      = 1 << 0,
+      TR_DROPPING     = 1 << 1,
+      TR_STORED_TABLE = 1 << 2
+    };
+    Uint8 get_enabled()     const { return (m_flags & TR_ENABLED)      != 0; }
+    Uint8 get_dropping()    const { return (m_flags & TR_DROPPING)     != 0; }
+    Uint8 get_storedTable() const { return (m_flags & TR_STORED_TABLE) != 0; }
+    void set_enabled(Uint8 f)     { f ? m_flags |= (Uint16)TR_ENABLED      : m_flags &= ~(Uint16)TR_ENABLED; }
+    void set_dropping(Uint8 f)    { f ? m_flags |= (Uint16)TR_DROPPING     : m_flags &= ~(Uint16)TR_DROPPING; }
+    void set_storedTable(Uint8 f) { f ? m_flags |= (Uint16)TR_STORED_TABLE : m_flags &= ~(Uint16)TR_STORED_TABLE; }
+
+    Uint8 noOfKeyAttr;
+    Uint8 hasCharAttr;
+    Uint8 noOfDistrKeys;
     
     bool checkTable(Uint32 schemaVersion) const {
-      return enabled && !dropping && (schemaVersion == currentSchemaVersion);
+      return get_enabled() && !get_dropping() && 
+	(table_version_major(schemaVersion) == table_version_major(currentSchemaVersion));
     }
-    
+
     Uint32 getErrorCode(Uint32 schemaVersion) const;
-    
+
     struct DropTable {
       Uint32 senderRef;
       Uint32 senderData;
@@ -1167,7 +1165,7 @@ public:
     Uint32 nextScan;
 
     // Length of expected attribute information
-    Uint32 scanAiLength;
+    union { Uint32 scanAiLength; Uint32 m_booked_fragments_count; };
 
     Uint32 scanKeyLen;
 
@@ -1334,11 +1332,11 @@ private:
   void execCOMMITCONF(Signal* signal);
   void execABORTCONF(Signal* signal);
   void execNODE_FAILREP(Signal* signal);
+  void execNODE_START_REP(Signal* signal);
   void execINCL_NODEREQ(Signal* signal);
   void execTIME_SIGNAL(Signal* signal);
   void execAPI_FAILREQ(Signal* signal);
   void execSCAN_HBREP(Signal* signal);
-  void execSET_VAR_REQ(Signal* signal);
 
   void execABORT_ALL_REQ(Signal* signal);
 
@@ -1422,7 +1420,7 @@ private:
 		   const UintR scanParallel, 
 		   const UintR noOprecPerFrag);
   void initScanfragrec(Signal* signal);
-  void releaseScanResources(ScanRecordPtr);
+  void releaseScanResources(ScanRecordPtr, bool not_started = false);
   ScanRecordPtr seizeScanrec(Signal* signal);
   void sendScanFragReq(Signal*, ScanRecord*, ScanFragRec*);
   void sendScanTabConf(Signal* signal, ScanRecordPtr);
@@ -1436,6 +1434,10 @@ private:
   void gcpTcfinished(Signal* signal);
   void handleGcp(Signal* signal);
   void hash(Signal* signal);
+  bool handle_special_hash(Uint32 dstHash[4], 
+			     Uint32* src, Uint32 srcLen, 
+			     Uint32 tabPtrI, bool distr);
+  
   void initApiConnect(Signal* signal);
   void initApiConnectRec(Signal* signal, 
 			 ApiConnectRecord * const regApiPtr,
@@ -1481,7 +1483,7 @@ private:
   void sendContinueTimeOutControl(Signal* signal, Uint32 TapiConPtr);
   void sendKeyinfo(Signal* signal, BlockReference TBRef, Uint32 len);
   void sendlqhkeyreq(Signal* signal, BlockReference TBRef);
-  void sendSystemError(Signal* signal);
+  void sendSystemError(Signal* signal, int line);
   void sendtckeyconf(Signal* signal, UintR TcommitFlag);
   void sendTcIndxConf(Signal* signal, UintR TcommitFlag);
   void unlinkApiConnect(Signal* signal);
@@ -1495,12 +1497,12 @@ private:
   void clearCommitAckMarker(ApiConnectRecord * const regApiPtr,
 			    TcConnectRecord * const regTcPtr);
   // Trigger and index handling
-  bool saveINDXKEYINFO(Signal* signal,
-                       TcIndexOperation* indexOp,
-                       const Uint32 *src, 
-                       Uint32 len);
+  int saveINDXKEYINFO(Signal* signal,
+                      TcIndexOperation* indexOp,
+                      const Uint32 *src, 
+                      Uint32 len);
   bool receivedAllINDXKEYINFO(TcIndexOperation* indexOp);
-  bool saveINDXATTRINFO(Signal* signal,
+  int saveINDXATTRINFO(Signal* signal,
                         TcIndexOperation* indexOp,
                         const Uint32 *src, 
                         Uint32 len);
@@ -1557,15 +1559,16 @@ private:
                             bool holdOperation = false);
   void releaseFiredTriggerData(DLFifoList<TcFiredTriggerData>* triggers);
   // Generated statement blocks
-  void warningHandlerLab(Signal* signal);
-  void systemErrorLab(Signal* signal);
+  void warningHandlerLab(Signal* signal, int line);
+  void systemErrorLab(Signal* signal, int line);
   void sendSignalErrorRefuseLab(Signal* signal);
   void scanTabRefLab(Signal* signal, Uint32 errCode);
   void diFcountReqLab(Signal* signal, ScanRecordPtr);
   void signalErrorRefuseLab(Signal* signal);
   void abort080Lab(Signal* signal);
   void packKeyData000Lab(Signal* signal, BlockReference TBRef, Uint32 len);
-  void abortScanLab(Signal* signal, ScanRecordPtr, Uint32 errCode);
+  void abortScanLab(Signal* signal, ScanRecordPtr, Uint32 errCode, 
+		    bool not_started = false);
   void sendAbortedAfterTimeout(Signal* signal, int Tcheck);
   void abort010Lab(Signal* signal);
   void abort015Lab(Signal* signal);
@@ -1603,7 +1606,7 @@ private:
   void wrongSchemaVersionErrorLab(Signal* signal);
   void noFreeConnectionErrorLab(Signal* signal);
   void tckeyreq050Lab(Signal* signal);
-  void timeOutFoundLab(Signal* signal, UintR anAdd);
+  void timeOutFoundLab(Signal* signal, UintR anAdd, Uint32 errCode);
   void completeTransAtTakeOverLab(Signal* signal, UintR TtakeOverInd);
   void completeTransAtTakeOverDoLast(Signal* signal, UintR TtakeOverInd);
   void completeTransAtTakeOverDoOne(Signal* signal, UintR TtakeOverInd);
@@ -1625,6 +1628,9 @@ private:
   void checkScanFragList(Signal*, Uint32 failedNodeId, ScanRecord * scanP, 
 			 LocalDLList<ScanFragRec>::Head&);
 
+  void nodeFailCheckTransactions(Signal*,Uint32 transPtrI,Uint32 failedNodeId);
+  void checkNodeFailComplete(Signal* signal, Uint32 failedNodeId, Uint32 bit);
+  
   // Initialisation
   void initData();
   void initRecords();
@@ -1651,6 +1657,7 @@ private:
   HostRecord *hostRecord;
   HostRecordPtr hostptr;
   UintR chostFilesize;
+  NdbNodeBitmask c_alive_nodes;
 
   GcpRecord *gcpRecord;
   GcpRecordPtr gcpPtr;
@@ -1671,6 +1678,7 @@ private:
   UintR tcheckGcpId;
 
   struct TransCounters {
+    TransCounters() {}
     enum { Off, Timer, Started } c_trans_status;
     UintR cattrinfoCount;
     UintR ctransCount;
@@ -1688,7 +1696,7 @@ private:
 	c_scan_count = c_range_scan_count = 0; 
     }
     Uint32 report(Signal* signal){
-      signal->theData[0] = EventReport::TransReportCounters;
+      signal->theData[0] = NDB_LE_TransReportCounters;
       signal->theData[1] = ctransCount;
       signal->theData[2] = ccommitCount;
       signal->theData[3] = creadCount;
@@ -1809,6 +1817,7 @@ private:
    */
 public:
   struct CommitAckMarker {
+    CommitAckMarker() {}
     Uint32 transid1;
     Uint32 transid2;
     union { Uint32 nextPool; Uint32 nextHash; };
@@ -1841,9 +1850,14 @@ private:
 			Uint32 transid2);
   void removeMarkerForFailedAPI(Signal* signal, Uint32 nodeId, Uint32 bucket);
 
-  bool getAllowStartTransaction() const {
-    if(getNodeState().getSingleUserMode())
-      return true;
+  bool getAllowStartTransaction(Uint32 nodeId, Uint32 table_single_user_mode) const {
+    if (unlikely(getNodeState().getSingleUserMode()))
+    {
+      if (getNodeState().getSingleUserApi() == nodeId || table_single_user_mode)
+        return true;
+      else
+        return false;
+    }
     return getNodeState().startLevel < NodeState::SL_STOPPING_2;
   }
   
@@ -1956,5 +1970,8 @@ private:
   // those variables should be removed and exchanged for stack
   // variable communication.
   /**************************************************************************/
+
+  Uint32 c_gcp_ref;
 };
+
 #endif
