@@ -2,8 +2,7 @@
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
-   the Free Software Foundation; either version 2 of the License, or
-   (at your option) any later version.
+   the Free Software Foundation; version 2 of the License.
 
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
@@ -305,6 +304,10 @@ Dbtup::primaryKey(Tablerec* const regTabPtr, Uint32 attrId)
 Uint32
 Dbtup::dropTrigger(Tablerec* table, const DropTrigReq* req)
 {
+  if (ERROR_INSERTED(4004)) {
+    CLEAR_ERROR_INSERT_VALUE;
+    return 9999;
+  }
   Uint32 triggerId = req->getTriggerId();
 
   TriggerType::Value ttype = req->getTriggerType();
@@ -603,7 +606,7 @@ void Dbtup::executeTrigger(Signal* signal,
     for everybody else.
     */
     signal->theData[0] = trigPtr->triggerId;
-    signal->theData[1] = regOperPtr->fragId;
+    signal->theData[1] = regOperPtr->fragId >> 1; // send "real" frag id
     EXECUTE_DIRECT(BACKUP, GSN_BACKUP_TRIG_REQ, signal, 2);
     ljamEntry();
     if (signal->theData[0] == 0) {
@@ -753,7 +756,7 @@ bool Dbtup::readTriggerInfo(TupTriggerData* const trigPtr,
 			  regTabPtr->noOfKeyAttr,
 			  keyBuffer,
 			  ZATTR_BUFFER_SIZE,
-			  true);
+			  false);
   ndbrequire(ret != -1);
   noPrimKey= ret;
 
@@ -796,7 +799,7 @@ bool Dbtup::readTriggerInfo(TupTriggerData* const trigPtr,
 			    numAttrsToRead,
 			    mainBuffer,
 			    ZATTR_BUFFER_SIZE,
-			    true);
+			    false);
     ndbrequire(ret != -1);
     noMainWords= ret;
   } else {
@@ -822,7 +825,7 @@ bool Dbtup::readTriggerInfo(TupTriggerData* const trigPtr,
 			    numAttrsToRead,
 			    copyBuffer,
 			    ZATTR_BUFFER_SIZE,
-			    true);
+			    false);
 
     ndbrequire(ret != -1);
     noCopyWords = ret;
