@@ -2,8 +2,7 @@
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
-   the Free Software Foundation; either version 2 of the License, or
-   (at your option) any later version.
+   the Free Software Foundation; version 2 of the License.
 
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
@@ -67,7 +66,6 @@ NdbOperation::equal_impl(const NdbColumnImpl* tAttrInfo,
   
   register Uint32 tAttrId;
 
-  Uint32 tData;
   Uint32 tKeyInfoPosition;
   const char* aValue = aValuePassed;
   Uint64 tempData[512];
@@ -347,7 +345,6 @@ NdbOperation::insertKEYINFO(const char* aValue,
   Uint32 tEndPos;
   Uint32 tPos;
   Uint32 signalCounter;
-  Uint32 tData;
 
 /*****************************************************************************
  *	Calculate the end position of the attribute in the key information.  *
@@ -514,7 +511,6 @@ NdbOperation::handle_distribution_key(const Uint64* value, Uint32 len)
      * Copy distribution key to linear memory
      */
     NdbColumnImpl* const * cols = m_accessTable->m_columns.getBase();
-    Uint32 len = 0;
     Uint64 tmp[1000];
 
     Uint32 chunk = 8;
