@@ -2,8 +2,7 @@
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
-   the Free Software Foundation; either version 2 of the License, or
-   (at your option) any later version.
+   the Free Software Foundation; version 2 of the License.
 
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
@@ -190,6 +189,7 @@ static struct my_option my_long_options[] =
 };
 static void usage()
 {
+#ifdef NOT_USED
   char desc[] = 
     "tabname\n"\
     "This program list all system objects in  NDB Cluster.\n"\
@@ -197,6 +197,7 @@ static void usage()
     " ex: ndb_show_tables -t 2 would show all UserTables\n"\
     "To show all indexes for a table write table name as final argument\n"\
     "  ex: ndb_show_tables T1\n";
+#endif
   ndb_std_print_version();
   print_defaults(MYSQL_CONFIG_NAME,load_default_groups);
   puts("");
@@ -218,6 +219,7 @@ int main(int argc, char** argv){
   _tabname = argv[0];
 
   ndb_cluster_connection = new Ndb_cluster_connection(opt_connect_str);
+  ndb_cluster_connection->set_name("ndb_show_tables");
   if (ndb_cluster_connection->connect(12,5,1))
     fatal("Unable to connect to management server.");
   if (ndb_cluster_connection->wait_until_ready(30,0) < 0)
