@@ -2,8 +2,7 @@
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
-   the Free Software Foundation; either version 2 of the License, or
-   (at your option) any later version.
+   the Free Software Foundation; version 2 of the License.
 
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
@@ -21,87 +20,102 @@
 #ifdef USE_MB
 
 
-void my_caseup_str_mb(CHARSET_INFO * cs, char *str)
+uint my_caseup_str_mb(CHARSET_INFO * cs, char *str)
 {
   register uint32 l;
-  register char *end=str+strlen(str); /* BAR TODO: remove strlen() call */
-  register uchar *map=cs->to_upper;
+  register uchar *map= cs->to_upper;
+  char *str_orig= str;
   
   while (*str)
   {
-    if ((l=my_ismbchar(cs, str,end)))
-      str+=l;
+    /* Pointing after the '\0' is safe here. */
+    if ((l= my_ismbchar(cs, str, str + cs->mbmaxlen)))
+      str+= l;
     else
     { 
-      *str=(char) map[(uchar)*str];
+      *str= (char) map[(uchar)*str];
       str++;
     }
   }
+  return str - str_orig;
 }
 
-void my_casedn_str_mb(CHARSET_INFO * cs, char *str)
+uint my_casedn_str_mb(CHARSET_INFO * cs, char *str)
 {
   register uint32 l;
-  register char *end=str+strlen(str);
-  register uchar *map=cs->to_lower;
+  register uchar *map= cs->to_lower;
+  char *str_orig= str;
   
   while (*str)
   {
-    if ((l=my_ismbchar(cs, str,end)))
-      str+=l;
+    /* Pointing after the '\0' is safe here. */
+    if ((l= my_ismbchar(cs, str, str + cs->mbmaxlen)))
+      str+= l;
     else
     {
-      *str=(char) map[(uchar)*str];
+      *str= (char) map[(uchar)*str];
       str++;
     }
   }
+  return str - str_orig;
 }
 
-void my_caseup_mb(CHARSET_INFO * cs, char *str, uint length)
+uint my_caseup_mb(CHARSET_INFO * cs, char *src, uint srclen,
+                  char *dst __attribute__((unused)),
+                  uint dstlen __attribute__((unused)))
 {
   register uint32 l;
-  register char *end=str+length;
-  register uchar *map=cs->to_upper;
-  
-  while (str<end)
+  register char *srcend= src + srclen;
+  register uchar *map= cs->to_upper;
+
+  DBUG_ASSERT(src == dst && srclen == dstlen);
+  while (src < srcend)
   {
-    if ((l=my_ismbchar(cs, str,end)))
-      str+=l;
+    if ((l=my_ismbchar(cs, src, srcend)))
+      src+= l;
     else 
     {
-      *str=(char) map[(uchar)*str];
-      str++;
+      *src=(char) map[(uchar) *src];
+      src++;
     }
   }
+  return srclen;
 }
 
-void my_casedn_mb(CHARSET_INFO * cs, char *str, uint length)
+uint my_casedn_mb(CHARSET_INFO * cs, char *src, uint srclen,
+                  char *dst __attribute__((unused)),
+                  uint dstlen __attribute__((unused)))
 {
   register uint32 l;
-  register char *end=str+length;
+  register char *srcend= src + srclen;
   register uchar *map=cs->to_lower;
-  
-  while (str<end)
+
+  DBUG_ASSERT(src == dst && srclen == dstlen);  
+  while (src < srcend)
   {
-    if ((l=my_ismbchar(cs, str,end)))
-      str+=l;
+    if ((l= my_ismbchar(cs, src, srcend)))
+      src+= l;
     else
     {
-      *str=(char) map[(uchar)*str];
-      str++;
+      *src= (char) map[(uchar)*src];
+      src++;
     }
   }
+  return srclen;
 }
 
+/*
+  my_strcasecmp_mb() returns 0 if strings are equal, non-zero otherwise.
+ */
 int my_strcasecmp_mb(CHARSET_INFO * cs,const char *s, const char *t)
 {
   register uint32 l;
-  register const char *end=s+strlen(s);
   register uchar *map=cs->to_upper;
   
-  while (s<end)
+  while (*s && *t)
   {
-    if ((l=my_ismbchar(cs, s,end)))
+    /* Pointing after the '\0' is safe here. */
+    if ((l=my_ismbchar(cs, s, s + cs->mbmaxlen)))
     {
       while (l--)
         if (*s++ != *t++) 
@@ -112,7 +126,8 @@ int my_strcasecmp_mb(CHARSET_INFO * cs,const char *s, const char *t)
     else if (map[(uchar) *s++] != map[(uchar) *t++])
       return 1;
   }
-  return *t;
+  /* At least one of '*s' and '*t' is zero here. */
+  return (*t != *s);
 }
 
 
@@ -170,7 +185,7 @@ int my_wildcmp_mb(CHARSET_INFO *cs,
     {						/* Found w_many */
       uchar cmp;
       const char* mb = wildstr;
-      int mblen=0;
+      int mb_len=0;
       
       wildstr++;
       /* Remove any '%' and '_' from the wild search string */
@@ -196,7 +211,7 @@ int my_wildcmp_mb(CHARSET_INFO *cs,
 	cmp= *++wildstr;
 	
       mb=wildstr;
-      mblen= my_ismbchar(cs, wildstr, wildend);
+      mb_len= my_ismbchar(cs, wildstr, wildend);
       INC_PTR(cs,wildstr,wildend);		/* This is compared trough cmp */
       cmp=likeconv(cs,cmp);   
       do
@@ -205,11 +220,11 @@ int my_wildcmp_mb(CHARSET_INFO *cs,
         {
           if (str >= str_end)
             return -1;
-          if (mblen)
+          if (mb_len)
           {
-            if (str+mblen <= str_end && memcmp(str, mb, mblen) == 0)
+            if (str+mb_len <= str_end && memcmp(str, mb, mb_len) == 0)
             {
-              str += mblen;
+              str += mb_len;
               break;
             }
           }
@@ -241,8 +256,8 @@ uint my_numchars_mb(CHARSET_INFO *cs __attribute__((unused)),
   register uint32 count=0;
   while (pos < end) 
   {
-    uint mblen;
-    pos+= (mblen= my_ismbchar(cs,pos,end)) ? mblen : 1;
+    uint mb_len;
+    pos+= (mb_len= my_ismbchar(cs,pos,end)) ? mb_len : 1;
     count++;
   }
   return count;
@@ -256,11 +271,11 @@ uint my_charpos_mb(CHARSET_INFO *cs __attribute__((unused)),
   
   while (length && pos < end)
   {
-    uint mblen;
-    pos+= (mblen= my_ismbchar(cs, pos, end)) ? mblen : 1;
+    uint mb_len;
+    pos+= (mb_len= my_ismbchar(cs, pos, end)) ? mb_len : 1;
     length--;
   }
-  return length ? (uint) (end + 2 - start) : (uint) (pos - start);
+  return (uint) (length ? end+2-start : pos-start);
 }
 
 
@@ -272,14 +287,14 @@ uint my_well_formed_len_mb(CHARSET_INFO *cs, const char *b, const char *e,
   while (pos)
   {
     my_wc_t wc;
-    int mblen;
+    int mb_len;
 
-    if ((mblen= cs->cset->mb_wc(cs, &wc, (uchar*) b, (uchar*) e)) <= 0)
+    if ((mb_len= cs->cset->mb_wc(cs, &wc, (uchar*) b, (uchar*) e)) <= 0)
     {
       *error= b < e ? 1 : 0;
       break;
     }
-    b+= mblen;
+    b+= mb_len;
     pos--;
   }
   return (uint) (b - b_start);
@@ -303,7 +318,7 @@ uint my_instr_mb(CHARSET_INFO *cs,
       {
         match->beg= 0;
         match->end= 0;
-        match->mblen= 0;
+        match->mb_len= 0;
       }
       return 1;		/* Empty string is always found */
     }
@@ -313,7 +328,7 @@ uint my_instr_mb(CHARSET_INFO *cs,
     
     while (b < end)
     {
-      int mblen;
+      int mb_len;
       
       if (!cs->coll->strnncoll(cs, (unsigned char*) b,   s_length, 
       				   (unsigned char*) s, s_length, 0))
@@ -321,20 +336,20 @@ uint my_instr_mb(CHARSET_INFO *cs,
         if (nmatch)
         {
           match[0].beg= 0;
-          match[0].end= b-b0;
-          match[0].mblen= res;
+          match[0].end= (uint) (b-b0);
+          match[0].mb_len= res;
           if (nmatch > 1)
           {
             match[1].beg= match[0].end;
             match[1].end= match[0].end+s_length;
-            match[1].mblen= 0;	/* Not computed */
+            match[1].mb_len= 0;	/* Not computed */
           }
         }
         return 2;
       }
-      mblen= (mblen= my_ismbchar(cs, b, end)) ? mblen : 1;
-      b+= mblen;
-      b_length-= mblen;
+      mb_len= (mb_len= my_ismbchar(cs, b, end)) ? mb_len : 1;
+      b+= mb_len;
+      b_length-= mb_len;
       res++;
     }
   }
@@ -365,6 +380,9 @@ static int my_strnncoll_mb_bin(CHARSET_INFO * cs __attribute__((unused)),
     slen		Length of 's'
     t			String to compare
     tlen		Length of 't'
+    diff_if_only_endspace_difference
+		        Set to 1 if the strings should be regarded as different
+                        if they only difference in end space
 
   NOTE
    This function is used for character strings with binary collations.
@@ -379,10 +397,16 @@ static int my_strnncoll_mb_bin(CHARSET_INFO * cs __attribute__((unused)),
 
 static int my_strnncollsp_mb_bin(CHARSET_INFO * cs __attribute__((unused)),
                                  const uchar *a, uint a_length, 
-                                 const uchar *b, uint b_length)
+                                 const uchar *b, uint b_length,
+                                 my_bool diff_if_only_endspace_difference)
 {
   const uchar *end;
   uint length;
+  int res;
+
+#ifndef VARCHAR_WITH_DIFF_ENDSPACE_ARE_DIFFERENT_FOR_UNIQUE
+  diff_if_only_endspace_difference= 0;
+#endif
   
   end= a + (length= min(a_length, b_length));
   while (a < end)
@@ -390,9 +414,12 @@ static int my_strnncollsp_mb_bin(CHARSET_INFO * cs __attribute__((unused)),
     if (*a++ != *b++)
       return ((int) a[-1] - (int) b[-1]);
   }
+  res= 0;
   if (a_length != b_length)
   {
     int swap= 1;
+    if (diff_if_only_endspace_difference)
+      res= 1;                                   /* Assume 'a' is bigger */
     /*
       Check the next not space character of the longer key. If it's < ' ',
       then it's smaller than the other key.
@@ -403,6 +430,7 @@ static int my_strnncollsp_mb_bin(CHARSET_INFO * cs __attribute__((unused)),
       a_length= b_length;
       a= b;
       swap= -1;					/* swap sign of result */
+      res= -res;
     }
     for (end= a + a_length-length; a < end ; a++)
     {
@@ -410,7 +438,7 @@ static int my_strnncollsp_mb_bin(CHARSET_INFO * cs __attribute__((unused)),
 	return (*a < ' ') ? -swap : swap;
     }
   }
-  return 0;
+  return res;
 }
 
 
@@ -439,6 +467,13 @@ static void my_hash_sort_mb_bin(CHARSET_INFO *cs __attribute__((unused)),
   
   key+= len;
   
+  /*
+     Remove trailing spaces. We have to do this to be able to compare
+    'A ' and 'A' as identical
+  */
+  while (key > pos && key[-1] == ' ')
+    key--;
+  
   for (; pos < (uchar*) key ; pos++)
   {
     nr1[0]^=(ulong) ((((uint) nr1[0] & 63)+nr2[0]) * 
@@ -449,15 +484,35 @@ static void my_hash_sort_mb_bin(CHARSET_INFO *cs __attribute__((unused)),
 
 
 /* 
-  Write max key: create a buffer with multibyte
-  representation of the max_sort_char character,
-  and copy it into max_str in a loop. 
+  Fill the given buffer with 'maximum character' for given charset
+  SYNOPSIS
+      pad_max_char()
+      cs   Character set
+      str  Start of buffer to fill
+      end  End of buffer to fill
+
+  DESCRIPTION
+      Write max key:
+      - for non-Unicode character sets:
+        just set to 255.
+      - for Unicode character set (utf-8):
+        create a buffer with multibyte representation of the max_sort_char
+        character, and copy it into max_str in a loop. 
 */
 static void pad_max_char(CHARSET_INFO *cs, char *str, char *end)
 {
   char buf[10];
-  char buflen= cs->cset->wc_mb(cs, cs->max_sort_char, (uchar*) buf,
-                               (uchar*) buf + sizeof(buf));
+  char buflen;
+  
+  if (!(cs->state & MY_CS_UNICODE))
+  {
+    bfill(str, end - str, 255);
+    return;
+  }
+  
+  buflen= cs->cset->wc_mb(cs, cs->max_sort_char, (uchar*) buf,
+                          (uchar*) buf + sizeof(buf));
+  
   DBUG_ASSERT(buflen > 0);
   do
   {
@@ -502,24 +557,32 @@ my_bool my_like_range_mb(CHARSET_INFO *cs,
 			 char *min_str,char *max_str,
 			 uint *min_length,uint *max_length)
 {
+  uint mb_len;
   const char *end= ptr + ptr_length;
   char *min_org= min_str;
   char *min_end= min_str + res_length;
   char *max_end= max_str + res_length;
-  uint charlen= res_length / cs->mbmaxlen;
+  uint maxcharlen= res_length / cs->mbmaxlen;
+  const char *contraction_flags= cs->contractions ? 
+              ((const char*) cs->contractions) + 0x40*0x40 : NULL;
 
-  for (; ptr != end && min_str != min_end && charlen > 0 ; ptr++, charlen--)
+  for (; ptr != end && min_str != min_end && maxcharlen ; maxcharlen--)
   {
+    /* We assume here that escape, w_any, w_namy are one-byte characters */
     if (*ptr == escape && ptr+1 != end)
-    {
-      ptr++;					/* Skip escape */
-      *min_str++= *max_str++ = *ptr;
-      continue;
-    }
-    if (*ptr == w_one || *ptr == w_many)	/* '_' and '%' in SQL */
-    {
-      /* Write min key  */
-      *min_length= (uint) (min_str - min_org);
+      ptr++;                                    /* Skip escape */
+    else if (*ptr == w_one || *ptr == w_many)   /* '_' and '%' in SQL */
+    {      
+fill_max_and_min:
+      /*
+        Calculate length of keys:
+        'a\0\0... is the smallest possible string when we have space expand
+        a\ff\ff... is the biggest possible string
+      */
+      *min_length= ((cs->state & MY_CS_BINSORT) ? (uint) (min_str - min_org) :
+                    res_length);
+      *max_length= res_length;
+      /* Create min key  */
       do
       {
 	*min_str++= (char) cs->min_sort_char;
@@ -534,16 +597,91 @@ my_bool my_like_range_mb(CHARSET_INFO *cs,
       pad_max_char(cs, max_str, max_end);
       return 0;
     }
-    *min_str++= *max_str++ = *ptr;
-  }
-  *min_length= *max_length = (uint) (min_str - min_org);
+    if ((mb_len= my_ismbchar(cs, ptr, end)) > 1)
+    {
+      if (ptr+mb_len > end || min_str+mb_len > min_end)
+        break;
+      while (mb_len--)
+       *min_str++= *max_str++= *ptr++;
+    }
+    else
+    {
+      /*
+        Special case for collations with contractions.
+        For example, in Chezh, 'ch' is a separate letter
+        which is sorted between 'h' and 'i'.
+        If the pattern 'abc%', 'c' at the end can mean:
+        - letter 'c' itself,
+        - beginning of the contraction 'ch'.
 
+        If we simply return this LIKE range:
+
+         'abc\min\min\min' and 'abc\max\max\max'
+
+        then this query: SELECT * FROM t1 WHERE a LIKE 'abc%'
+        will only find values starting from 'abc[^h]',
+        but won't find values starting from 'abch'.
+
+        We must ignore contraction heads followed by w_one or w_many.
+        ('Contraction head' means any letter which can be the first
+        letter in a contraction)
+
+        For example, for Czech 'abc%', we will return LIKE range,
+        which is equal to LIKE range for 'ab%':
+
+        'ab\min\min\min\min' and 'ab\max\max\max\max'.
+
+      */
+      if (contraction_flags && ptr + 1 < end &&
+          contraction_flags[(uchar) *ptr])
+      {
+        /* Ptr[0] is a contraction head. */
+        
+        if (ptr[1] == w_one || ptr[1] == w_many)
+        {
+          /* Contraction head followed by a wildcard, quit. */
+          goto fill_max_and_min;
+        }
+        
+        /*
+          Some letters can be both contraction heads and contraction tails.
+          For example, in Danish 'aa' is a separate single letter which
+          is sorted after 'z'. So 'a' can be both head and tail.
+          
+          If ptr[0]+ptr[1] is a contraction,
+          then put both letters together.
+          
+          If ptr[1] can be a contraction part, but ptr[0]+ptr[1]
+          is not a contraction, then we put only ptr[0],
+          and continue with ptr[1] on the next loop.
+        */
+        if (contraction_flags[(uchar) ptr[1]] &&
+            cs->contractions[(*ptr-0x40)*0x40 + ptr[1] - 0x40])
+        {
+          /* Contraction found */
+          if (maxcharlen == 1 || min_str + 1 >= min_end)
+          {
+            /* Both contraction parts don't fit, quit */
+            goto fill_max_and_min;
+          }
+          
+          /* Put contraction head */
+          *min_str++= *max_str++= *ptr++;
+          maxcharlen--;
+        }
+      }
+      /* Put contraction tail, or a single character */
+      *min_str++= *max_str++= *ptr++;    
+    }
+  }
+
+  *min_length= *max_length = (uint) (min_str - min_org);
   while (min_str != min_end)
-    *min_str++= ' ';		/* Because if key compression */
-  pad_max_char(cs, max_str, max_end);
+    *min_str++= *max_str++= ' ';           /* Because if key compression */
   return 0;
 }
 
+
 static int my_wildcmp_mb_bin(CHARSET_INFO *cs,
 		  const char *str,const char *str_end,
 		  const char *wildstr,const char *wildend,
@@ -587,7 +725,7 @@ static int my_wildcmp_mb_bin(CHARSET_INFO *cs,
     {						/* Found w_many */
       uchar cmp;
       const char* mb = wildstr;
-      int mblen=0;
+      int mb_len=0;
       
       wildstr++;
       /* Remove any '%' and '_' from the wild search string */
@@ -613,7 +751,7 @@ static int my_wildcmp_mb_bin(CHARSET_INFO *cs,
 	cmp= *++wildstr;
 	
       mb=wildstr;
-      mblen= my_ismbchar(cs, wildstr, wildend);
+      mb_len= my_ismbchar(cs, wildstr, wildend);
       INC_PTR(cs,wildstr,wildend);		/* This is compared trough cmp */
       do
       {
@@ -621,11 +759,11 @@ static int my_wildcmp_mb_bin(CHARSET_INFO *cs,
         {
           if (str >= str_end)
             return -1;
-          if (mblen)
+          if (mb_len)
           {
-            if (str+mblen <= str_end && memcmp(str, mb, mblen) == 0)
+            if (str+mb_len <= str_end && memcmp(str, mb, mb_len) == 0)
             {
-              str += mblen;
+              str += mb_len;
               break;
             }
           }
@@ -864,15 +1002,15 @@ uint my_numcells_mb(CHARSET_INFO *cs, const char *b, const char *e)
   
   while (b < e)
   {
-    int mblen;
+    int mb_len;
     uint pg;
-    if ((mblen= cs->cset->mb_wc(cs, &wc, (uchar*) b, (uchar*) e)) <= 0)
+    if ((mb_len= cs->cset->mb_wc(cs, &wc, (uchar*) b, (uchar*) e)) <= 0)
     {
-      mblen= 1; /* Let's think a wrong sequence takes 1 dysplay cell */
+      mb_len= 1; /* Let's think a wrong sequence takes 1 dysplay cell */
       b++;
       continue;
     }
-    b+= mblen;
+    b+= mb_len;
     pg= (wc >> 8) & 0xFF;
     clen+= utr11_data[pg].p ? utr11_data[pg].p[wc & 0xFF] : utr11_data[pg].page;
     clen++;
@@ -887,11 +1025,13 @@ MY_COLLATION_HANDLER my_collation_mb_bin_handler =
     my_strnncoll_mb_bin,
     my_strnncollsp_mb_bin,
     my_strnxfrm_mb_bin,
-    my_like_range_simple,
+    my_strnxfrmlen_simple,
+    my_like_range_mb,
     my_wildcmp_mb_bin,
     my_strcasecmp_mb_bin,
     my_instr_mb,
-    my_hash_sort_mb_bin
+    my_hash_sort_mb_bin,
+    my_propagate_simple
 };
 
 
