@@ -2,8 +2,7 @@
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
-   the Free Software Foundation; either version 2 of the License, or
-   (at your option) any later version.
+   the Free Software Foundation; version 2 of the License.
 
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
@@ -180,8 +179,8 @@ int my_strnncollsp_simple(CHARSET_INFO * cs, const uchar *a, uint a_length,
     }
     for (end= a + a_length-length; a < end ; a++)
     {
-      if (*a != ' ')
-	return (*a < ' ') ? -swap : swap;
+      if (map[*a] != ' ')
+	return (map[*a] < ' ') ? -swap : swap;
     }
   }
   return res;
@@ -1174,7 +1173,7 @@ uint my_instr_simple(CHARSET_INFO *cs,
       {
         match->beg= 0;
         match->end= 0;
-        match->mblen= 0;
+        match->mb_len= 0;
       }
       return 1;		/* Empty string is always found */
     }
@@ -1202,13 +1201,13 @@ skip:
 	{
 	  match[0].beg= 0;
 	  match[0].end= (uint) (str- (const uchar*)b-1);
-	  match[0].mblen= match[0].end;
+	  match[0].mb_len= match[0].end;
 	  
 	  if (nmatch > 1)
 	  {
 	    match[1].beg= match[0].end;
 	    match[1].end= match[0].end+s_length;
-	    match[1].mblen= match[1].end-match[1].beg;
+	    match[1].mb_len= match[1].end-match[1].beg;
 	  }
 	}
 	return 2;
@@ -1586,19 +1585,19 @@ exp:    /* [ E [ <sign> ] <unsigned integer> ] */
     str++;
     if (str < end)
     {
-      int negative_exp, exp;
+      int negative_exp, exponent;
       if ((negative_exp= (*str == '-')) || *str=='+')
       {
         if (++str == end)
           goto ret_sign;
       }
-      for (exp= 0 ;
+      for (exponent= 0 ;
            str < end && (ch= (unsigned char) (*str - '0')) < 10;
            str++)
       {
-        exp= exp * 10 + ch;
+        exponent= exponent * 10 + ch;
       }
-      shift+= negative_exp ? -exp : exp;
+      shift+= negative_exp ? -exponent : exponent;
     }
   }
   
